!#########################################################################
!  Crystal group Routines for MkInvar: 
!     GenGroup      -- Generates space group for cells.
!     GenGroup_Unit -- Use generators of group to generate the
!                       symmetry group for the unit cell: integer.
!     GenGroup_Sim  -- Use symmetry group of unit cell to generate
!                       group for simulation cell which has the
!                       same symmetry as the unit cell: integer
!     Sym_Compare   -- Compare symmetry element with list of 
!                       symmetry elements and test if unique.
!
!#########################################################################    
      subroutine GenGroup()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer i,d
c
      call GenGroup_Unit()
c
c -- Simulation cell larger than unit cell w/ symmetry of unit cell
      d=0
      if(SIMCELL) d=SimCellDim(1)
      if(COVERCELL) d=CovCellDim
      if(d.ne.0) call GenGroup_Sim(d)
c
      return
      end
c
c
      subroutine GenGroup_Unit()
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
c
      integer i,i1,i2,i3,iop,iNew,j,k,Trans(3,4)
      real*8 Prim(3,3),PrimInv(3,3)
      logical test
c
      if(DEBUG) call DEBUG_OC(0,'GenGroup_Unit',13)
c
c -- Initialization
      nSym=nGen                 ! Initial dimension of group
      do i=1,nSym
         do j=1,4
            do k=1,3
               Sym(k,j,i)=Gen(k,j,i)
            end do
         end do
      end do
c
c -- Cell Vectors for unit cell
c -- Used in case cell vectors were defined in input file for 
c -- non-primitive cell.
      do i=1,3
         do j=1,3
            if(i.eq.j) then
               Prim(i,j)=dfloat(LCD_VS)
               PrimInv(i,j)=1.d0/dfloat(LCD_VS)
            else
               Prim(i,j)=0.d0
               PrimInv(i,j)=0.d0
            end if
         end do
      end do
c
c -- We continue to multiply symmetry elements until the group stops growing
      iNew=nSym                 ! Length of new group.
      do while (.true.)
         nSym=iNew
         if(DEBUG) write(LOGF,*) 'Current Size of Group: ',nSym
c
c -- Looping over all pairs of symmetry elements in the current group, multiply
c -- and test whether the new element is already in the group.
         do i1=1,nSym
            do i2=1,nSym
               call Sym_Multiply(Sym(1,1,i1),Sym(1,1,i2),
     $              Sym(1,1,iNew+1),Prim,PrimInv)
c
c -- Test if the resulting element is already in the group
               call Sym_Compare(iNew+1,test)
               if(test) iNew=iNew+1 ! Add element to the list
c
c -- Enough memory allocated?
               if(iNew.gt.nSymMax) then
                  write(LOGF,*) 'Error: GenGroup_Unit'
                  write(LOGF,*) 'Not enough memory allocated'
                  write(LOGF,*) 'Requested(iNew):    ',iNew
                  write(LOGF,*) 'Allowed(nSymMax):   ',nSymMax
                  stop
               end if
            end do
         end do
c
c -- Has group stopped growing?
         if(nSym.eq.iNew) goto 90
         nSym=iNew
      end do                    ! end while (.true.)
 90   continue                  ! Group stopped growing
c
c -- Do we have the expected number of symmetry elements?
      if(DEBUG) then
         write(LOGF,*) 'Current Size of Group: ',nSym
         write(LOGF,*) 'Group has stopped growing.'
         write(LOGF,*)
         write(LOGF,*) '# of symmetry elements expected: ',nSymUnit
         write(LOGF,*) '# of symmetry elements found:    ',nSym
      end if
c
c -- Do we have the correct # of symmetry elements?
      if(nSymUnit.ne.nSym) then
         write(LOGF,*) 'Error: GenGroup_Unit'
         write(LOGF,*) '# of symmetry elements expected(nSymUnit): ',
     $        nSymUnit
         write(LOGF,*) '# of symmetry elements found(nSym):        ',
     $        nSym
         stop
      end if
c
c -- If this is a non-primitive unit cell, we have to apply the 
c -- translation group for a larger covering cell to generate
c -- the full group.
      if(.not.PRIMITIVE) then
         if(DEBUG) then 
            write(LOGF,*)
            write(LOGF,*) 'Generating Group for non-primitive cell'
         end if
c
c -- Initialize rotation(identity) matrix for translations
         Trans(1,1)=1
         Trans(1,2)=0
         Trans(1,3)=0
c
         Trans(2,1)=0
         Trans(2,2)=1
         Trans(2,3)=0
c
         Trans(3,1)=0
         Trans(3,2)=0
         Trans(3,3)=1
c
c -- Loop over unit cell symmetry elements and wrap translational part
c -- into non-primitive cell.
         Trans(1,4)=0
         Trans(2,4)=0
         Trans(3,4)=0
         iNew=0
         do iop=1,nSymUnit
            call Sym_Multiply(Sym(1,1,iop),Trans,Sym(1,1,iNew+1),Cell,
     $           CellInv)
            call Sym_Compare(iNew+1,test) ! We test in case the
            if(test) iNew=iNew+1          ! group gets smaller.
         end do                           ! I don't think it should...
         nSymUnit=iNew          ! # of symmetry elements in non-primitive cell
c
         if(DEBUG) then
            write(LOGF,*)
            write(LOGF,*) 'Wrap symmetry elements to non-primitive cell'
            write(LOGF,*) 'Size of Group: ',nSymUnit
            write(LOGF,*)
            write(LOGF,*) 'Apply translation group of covering cell'
         end if
c
c -- Loop over all translations contained in covering cell to 
c -- generate the complete group.
         do i1=NonPrimRange(1,1),NonPrimRange(2,1) ! Translations along lattice vector #1
            Trans(1,4)=i1*LCD_VS
            do i2=NonPrimRange(1,2),NonPrimRange(2,2) ! Translations along #2
               Trans(2,4)=i2*LCD_VS
               do i3=NonPrimRange(1,3),NonPrimRange(2,3) ! Translations along #3
                  if((i1.eq.i2).and.(i2.eq.i3).and.(i3.eq.0)) goto 100 ! skip over unit cell at origin
                  Trans(3,4)=i3*LCD_VS
c     
c -- Loop over symmetry operations of unit cell
                  do iop=1,nSymUnit 
                     call Sym_Multiply(Sym(1,1,iop),Trans,
     $                    Sym(1,1,nSym+1),Cell,CellInv)
                     call Sym_Compare(nSym+1,test)
                     if(test) nSym=nSym+1 ! Add element to the list
c
c -- Enough memory allocated?
                     if(nSym.gt.nSymMax) then
                        write(LOGF,*) 'Error: GenGroup_Unit'
                        write(LOGF,*) 'Not enough memory allocated'
                        write(LOGF,*) 'Requested(nSym):    ',nSym
                        write(LOGF,*) 'Allowed(nSymMax):   ',nSymMax
                        stop
                     end if
                  end do
c     
c -- We skip over unit cell at origin
 100              continue
               end do
            end do
         end do
c
         if(DEBUG) write(LOGF,*) 'Size of Group: ',nSym
      end if
      nSymUnit=nSym
c
c -- Write group to file
      open(unit=1,file='Group_Unit.txt')
      call Write_Group(1,LCD_VS)
      close(1)
c
      if(DEBUG) then 
         write(LOGF,*) 
         write(LOGF,*) '    Group written to Group_Unit.txt.'
         call DEBUG_OC(1,'GenGroup_Unit',13)
      end if
c
      return
      end
c
c
      subroutine GenGroup_Sim(CellDim)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
c
      integer i,i1,i2,i3,iop,j,CellDim,Trans(3,4),Range(2,3)
      real*8 Cell_Temp(3,3),CellInv_Temp(3,3)
      logical test
c
      if(DEBUG) then 
         call DEBUG_OC(0,'GenGroup_Sim',12)
         write(LOGF,*) 'Cell Dimensions: ',CellDim,CellDim,CellDim
      end if
c
c -- Has enough memory been allocated?
      i=(CellDim**3)*nSymUnit
      if(DEBUG.and.PRIMITIVE) then 
         write(LOGF,*)
         write(LOGF,*) 'Expected size of group in simulation cell: ',i
         write(LOGF,*) ' -- Not accurate for unit cells that do not',
     $        '    possess the full symmetry of the lattice.'
      end if
c
      if(i.gt.nSymMax) then
         write(LOGF,*)
         write(LOGF,*) 'Error: GenGroup_Sim_Sym'
         write(LOGF,*) 'Not enough memory allocated'
         write(LOGF,*) 'Symmetric simulation cell dim:    ',CellDim
         write(LOGF,*) 'Size of group in simulation cell: ',i
         write(LOGF,*) 'Allowed(nSymMax):                 ',nSymMax
         stop
      end if
c
c -- Generate cell vectors for larger cell which are also
c -- scaled by LCD_VS
      do i=1,3
         do j=1,3
            Cell_Temp(i,j)=Cell(i,j)*dfloat(CellDim)
            CellInv_Temp(i,j)=CellInv(i,j)/dfloat(CellDim)
         end do
      end do
c
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) 'Matrix of Lattice Vectors for Simulation Cell:'
         write(LOGF,10) (Cell_Temp(1,j)/dfloat(LCD_VS),j=1,3)
         write(LOGF,10) (Cell_Temp(2,j)/dfloat(LCD_VS),j=1,3)
         write(LOGF,10) (Cell_Temp(3,j)/dfloat(LCD_VS),j=1,3)
         write(LOGF,*)
         write(LOGF,*) 'Inverse of Matrix of Lattice Vectors:'
         write(LOGF,10) (CellInv_Temp(1,j)*dfloat(LCD_VS),j=1,3)
         write(LOGF,10) (CellInv_Temp(2,j)*dfloat(LCD_VS),j=1,3)
         write(LOGF,10) (CellInv_Temp(3,j)*dfloat(LCD_VS),j=1,3)
      end if
c
c -- Initialize rotation(identity) matrix for translations
      Trans(1,1)=1
      Trans(1,2)=0
      Trans(1,3)=0
c
      Trans(2,1)=0
      Trans(2,2)=1
      Trans(2,3)=0
c
      Trans(3,1)=0
      Trans(3,2)=0
      Trans(3,3)=1
c
c -- Identify range of indices for translations
      if(PRIMITIVE) then
         Range(1,1)=0
         Range(2,1)=CellDim-1
         Range(1,2)=0
         Range(2,2)=CellDim-1
         Range(1,3)=0
         Range(2,3)=CellDim-1
c
      else
c
         Range(1,1)=NonPrimRange(1,1)*CellDim
         Range(2,1)=NonPrimRange(2,1)*CellDim
         Range(1,2)=NonPrimRange(1,2)*CellDim
         Range(2,2)=NonPrimRange(2,2)*CellDim
         Range(1,3)=NonPrimRange(1,3)*CellDim
         Range(2,3)=NonPrimRange(2,3)*CellDim
      end if
c
c -- Loop over all unit cells(translations) contained in simulation cell
      do i1=Range(1,1),Range(2,1) ! Translations along lattice vector #1
         Trans(1,4)=i1*LCD_VS
         do i2=Range(1,2),Range(2,2) ! Translations along #2
            Trans(2,4)=i2*LCD_VS
            do i3=Range(1,3),Range(2,3) ! Translations along #3
               if((i1.eq.i2).and.(i2.eq.i3).and.(i3.eq.0)) goto 100 ! skip over unit cell at origin
               Trans(3,4)=i3*LCD_VS
c
c -- Loop over symmetry operations of unit cell
               do iop=1,nSymUnit 
                  call Sym_Multiply(Sym(1,1,iop),Trans,Sym(1,1,nSym+1),
     $                 Cell_Temp,CellInv_Temp)
                  call Sym_Compare(nSym+1,test)
                  if(test) nSym=nSym+1
c
c -- Enough memory allocated?
                  if(nSym.gt.nSymMax) then
                     write(LOGF,*) 'Error: GenGroup_Sim'
                     write(LOGF,*) 'Not enough memory allocated'
                     write(LOGF,*) 'Requested(nSym):    ',nSym
                     write(LOGF,*) 'Allowed(nSymMax):   ',nSymMax
                     stop
                  end if
               end do
c
c -- We skip over unit cell at origin
 100           continue
            end do
         end do
      end do
c
      if(DEBUG) then 
         write(LOGF,*)
         write(LOGF,*) 'Size of Group: ',nSym
         write(LOGF,*) 'Group written to Group_Sim.txt'
      end if
c
      open(unit=1,file='Group_Sim.txt')
      call Write_Group(1,LCD_VS)
      close(1)
c
      if(DEBUG) call DEBUG_OC(1,'GenGroup_Sim',12)
c
 10   format(3(1x,f5.2))
c
      return
      end
c
c
      subroutine Sym_Compare(n,test)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
c
      integer i,j,j1,k,n
      logical test
c
c -- Is the last element already present in the list?
      do i=1,n-1
         do j=1,4               ! The denominator is defined to be one whenever this routine is called.
            do k=1,3
               j1=abs(Sym(k,j,i)-Sym(k,j,n))
               if(j1.gt.0) goto 91
            end do
         end do
         goto 90                ! Symmetry elements are identical
c
 91      continue ! Symmetry elements are different
      end do
c
      test=.true.               ! The symmetry element is unique to this list
      return
c
 90   continue
      test=.false.              ! The symmetry element is already present in this list
c
      return
      end
