!#########################################################################
!  General Routines for MkInvar: 
!     Matrix_Transpose           -- Transpose of [3x3] matrix.
!     Matrix_Inverse             -- Inverse of [3x3] matrix.
!     Matrix_Multiply_Mv         -- Matrix-Vector product: [3x3].[3x1]
!     Matrix_Multiply_Mv_Integer --  "  " for integers
!     Matrix_Multiply_vM         -- Matrix-Vector product: [1x3].[3x3]
!     Matrix_Multiply_MN         -- Matrix-Matrix product: [3x3].[3x3]
!     Matrix_Multiply_MN_Integer --  "  " for integers
!     Mod_Positive               -- Return fractional part of real*8 number
!                                    If negative, add one.
!     Fraction_Reduce            -- Factor a rational number.
!     Prime_Factor               -- Factor integer into primes
!     Lowest_Common_Multiple     -- Find LCM for list of integers.
!     INTEGERtoSTRING            -- Convert integer to string of length 3
!                                    w/ leading zeroes.
!     List_Sort_Integer          -- Sort a list of integers in place
!     List_Union_Integer         -- Give list of unique elements in list
!     List_Min_Real              -- Give minimum element in list
!
!     DEBUG_OC                   -- Write output when entering/exiting 
!                                    subroutines
!     Check_File                 -- Determine if file exists
!     FUNC_NORM                  -- Norm of real*8 [3x1] vector
!     FUNC_DIST                  -- Distance between two real*8 [3x1] vectors.
!     EXIT_TERMINATE             -- Controlled termination of program
!     MemoryEstimate             -- Estimate for memory usage(.inc files)
!
!
!#########################################################################
      subroutine Matrix_Transpose(M)
      implicit none
c
      integer i,j
      real*8 M(3,3),N(3,3)
c
      do i=1,3
         do j=1,3
            N(i,j)=M(j,i)
         end do
      end do
c
      do i=1,3
         do j=1,3
            M(i,j)=N(i,j)
         end do
      end do
c
      return
      end
c
c
      subroutine Matrix_Inverse(A,I,LOGF)
      implicit none
c
      integer LOGF
      real*8 A(3,3),I(3,3),det,tol
      parameter(tol=0.00001d0)
c
c -- Calculate determinant of 3x3 matrix
      det=A(1,1)*(A(2,2)*A(3,3)-A(2,3)*A(3,2))-
     $     A(1,2)*(A(2,1)*A(3,3)-A(2,3)*A(3,1))+
     $     A(1,3)*(A(2,1)*A(3,2)-A(2,2)*A(3,1))
c
c -- Is the matrix close to singular?
      if(abs(det).le.tol) then
         write(LOGF,*) 'Error: MatrixInverse\n '
         write(LOGF,*) '3x3 matrix is close to singular(det): ',det
         stop
      end if
c
c -- Construct elements of the matrix inverse
      I(1,1)=(A(2,2)*A(3,3)-A(2,3)*A(3,2)) / det
      I(1,2)=(A(1,3)*A(3,2)-A(1,2)*A(3,3)) / det
      I(1,3)=(A(1,2)*A(2,3)-A(1,3)*A(2,2)) / det
c
      I(2,1)=(A(2,3)*A(3,1)-A(2,1)*A(3,3)) / det
      I(2,2)=(A(1,1)*A(3,3)-A(1,3)*A(3,1)) / det
      I(2,3)=(A(1,3)*A(2,1)-A(1,1)*A(2,3)) / det
c
      I(3,1)=(A(2,1)*A(3,2)-A(2,2)*A(3,1)) / det
      I(3,2)=(A(1,2)*A(3,1)-A(1,1)*A(3,2)) / det
      I(3,3)=(A(1,1)*A(2,2)-A(1,2)*A(2,1)) / det
c
      return
      end
c
c
      subroutine Matrix_Multiply_Mv(M,v,r)
      implicit none
c
      real*8 M(3,3),v(3),r(3)
c
      r(1)=M(1,1)*v(1)+M(1,2)*v(2)+M(1,3)*v(3)
      r(2)=M(2,1)*v(1)+M(2,2)*v(2)+M(2,3)*v(3)
      r(3)=M(3,1)*v(1)+M(3,2)*v(2)+M(3,3)*v(3)
c
      return
      end
c
c
      subroutine Matrix_Multiply_Mv_Integer(M,v,r)
      implicit none
c
      integer M(3,3),v(3),r(3)
c
      r(1)=M(1,1)*v(1)+M(1,2)*v(2)+M(1,3)*v(3)
      r(2)=M(2,1)*v(1)+M(2,2)*v(2)+M(2,3)*v(3)
      r(3)=M(3,1)*v(1)+M(3,2)*v(2)+M(3,3)*v(3)
c
      return
      end
c
c
      subroutine Matrix_Multiply_vM(v,M,r)
      implicit none
c
      real*8 M(3,3),v(3),r(3)
c
      r(1)=v(1)*M(1,1)+v(2)*M(2,1)+v(3)*M(3,1)
      r(2)=v(1)*M(1,2)+v(2)*M(2,2)+v(3)*M(3,2)
      r(3)=v(1)*M(1,3)+v(2)*M(2,3)+v(3)*M(3,3)
c
      return
      end
c
c
      subroutine Matrix_Multiply_MN(M,N,C)
      implicit none
c
      real*8 M(3,3),N(3,3),C(3,3)
c
c -- Column 1
      C(1,1)=M(1,1)*N(1,1)+M(1,2)*N(2,1)+M(1,3)*N(3,1)
      C(2,1)=M(2,1)*N(1,1)+M(2,2)*N(2,1)+M(2,3)*N(3,1)
      C(3,1)=M(3,1)*N(1,1)+M(3,2)*N(2,1)+M(3,3)*N(3,1)
c
c -- Column 2
      C(1,2)=M(1,1)*N(1,2)+M(1,2)*N(2,2)+M(1,3)*N(3,2)
      C(2,2)=M(2,1)*N(1,2)+M(2,2)*N(2,2)+M(2,3)*N(3,2)
      C(3,2)=M(3,1)*N(1,2)+M(3,2)*N(2,2)+M(3,3)*N(3,2)
c
c -- Column 3
      C(1,3)=M(1,1)*N(1,3)+M(1,2)*N(2,3)+M(1,3)*N(3,3)
      C(2,3)=M(2,1)*N(1,3)+M(2,2)*N(2,3)+M(2,3)*N(3,3)
      C(3,3)=M(3,1)*N(1,3)+M(3,2)*N(2,3)+M(3,3)*N(3,3)
c
      return
      end
c
c
      subroutine Matrix_Multiply_MN_Integer(M,N,C)
      implicit none
c
      integer M(3,3),N(3,3),C(3,3)
c
c -- Column 1
      C(1,1)=M(1,1)*N(1,1)+M(1,2)*N(2,1)+M(1,3)*N(3,1)
      C(2,1)=M(2,1)*N(1,1)+M(2,2)*N(2,1)+M(2,3)*N(3,1)
      C(3,1)=M(3,1)*N(1,1)+M(3,2)*N(2,1)+M(3,3)*N(3,1)
c
c -- Column 2
      C(1,2)=M(1,1)*N(1,2)+M(1,2)*N(2,2)+M(1,3)*N(3,2)
      C(2,2)=M(2,1)*N(1,2)+M(2,2)*N(2,2)+M(2,3)*N(3,2)
      C(3,2)=M(3,1)*N(1,2)+M(3,2)*N(2,2)+M(3,3)*N(3,2)
c
c -- Column 3
      C(1,3)=M(1,1)*N(1,3)+M(1,2)*N(2,3)+M(1,3)*N(3,3)
      C(2,3)=M(2,1)*N(1,3)+M(2,2)*N(2,3)+M(2,3)*N(3,3)
      C(3,3)=M(3,1)*N(1,3)+M(3,2)*N(2,3)+M(3,3)*N(3,3)
c
      return
      end
c
c
      real*8 function Mod_Positive(r)
      implicit none
c
      include 'param.inc'
c
      integer i
      real*8 r,r1
c
c -- This routine mimics the Mod[x,1] function in Mathematica.
c -- It returns the fractional part as a positive real number.
c -- For example, Mod[0.9,1] = 0.9 while Mod[-0.9,1] = 0.1.
c -- Mod[x,d] = x - d * Quotient[x,d]
c
      r1=r
c
c -- Is r within tolerance of being an integer
      i=int(r)
      if(abs(dfloat(i)-r).lt.TOL_ZERO) then
         r1=dfloat(i)
      elseif(abs(dfloat(i+1)-r).lt.TOL_ZERO) then
         r1=dfloat(i+1)
      elseif(abs(dfloat(i-1)-r).lt.TOL_ZERO) then
         r1=dfloat(i-1)
      end if
c
c -- Keep the fractional part
      Mod_Positive=r1-dfloat(int(r1))
c
c -- If negative, add one to make it positive
      if(Mod_Positive.lt.0.d0) Mod_Positive=Mod_Positive+1.d0
c
      return
      end
c
c
      subroutine Fraction_Reduce(n,d)
      implicit none
c
      integer n,d,l,r,sign
c
      sign=1
      if(n.lt.0) sign=-1
      n=abs(n)
c
      if(n.eq.0) then
         n=0
         d=1
      elseif(n.eq.d) then
         n=1
         d=1
      elseif(d.eq.1) then
         n=n
         d=d
      else
         if(n.gt.d) then
            l=d
         else
            l=n
         end if
c
 10      continue
         if(l.eq.1) goto 11
         do r=2,l
            if(((n/r)*r.eq.n).and.((d/r)*r.eq.d)) then
               n=n/r
               d=d/r
               l=l/r
               goto 10
            end if
         end do         
 11      continue
      end if
c
      n=n*sign
c
      return
      end
c
c
      subroutine Prime_Factor(n,expon)
      implicit none
c
      include 'param.inc'
c
      integer d,i,n,n1,expon(maxPrime)
c
c -- Initialization
      n1=n
      do i=1,maxPrime
         expon(i)=0
      end do
c
 10   continue
      do d=2,n1
         if((n1/d)*d.eq.n1) then
            if(d.gt.maxPrime) then
               write(LOGF,*) 'Error: Prime_Factor'
               write(LOGF,*) 'Prime too large:     ',d
               write(LOGF,*) 'Expecting(maxPrime): ',maxPrime
               stop
            end if
            n1=n1/d
            expon(d)=expon(d)+1 ! Exponent of primes
            if(d.gt.expon(1)) expon(1)=d ! Largest nonzero prime
            goto 10
         end if
      end do
c
      return
      end
c
c
      subroutine Lowest_Common_Multiple(n,dim,LCM)
      implicit none
c
      include 'param.inc'
c
      integer dim
      integer i,j,n(dim),le,lp,LCM,expon(maxPrime,dim)
c
c -- Is there only one entry in the list?
      if(dim.eq.1) then
         LCM=n(1)
      else
c
c -- Factor all integers into primes
         do i=1,dim
            call Prime_Factor(n(i),expon(1,i))
         end do
c
c -- What is the largest prime present?
         lp=expon(1,1)
         do i=2,dim
            if(expon(1,i).gt.lp) lp=expon(1,i)
         end do
c     
         LCM=1
         do i=lp,2,-1
c     
c -- Identifying largest exponent
            le=expon(i,1)
            do j=2,dim
               if(expon(i,j).gt.le) le=expon(i,j)
            end do
c
c -- Add contribution to LCM
            if(le.ne.0) LCM=LCM*(i**le)
         end do
      end if
c
      return
      end
c
c
      subroutine Greatest_Common_Divisor(n,dim,GCD)
      implicit none
c
      include 'param.inc'
c
      integer dim
      integer i,j,n(dim),GCD,lp,ge,expon(maxPrime,dim)
c
      if(dim.eq.1) then
         GCD=n(1)
      else
c
c -- Factor all integers into primes
         do i=1,dim
            call Prime_Factor(n(i),expon(1,i))
         end do
c
c -- What is the largest prime present?
         lp=expon(1,1)
         do i=2,dim
            if(expon(1,i).gt.lp) lp=expon(1,i)
         end do
c
         GCD=1
         do i=2,lp
c
c -- Identify greatest common exponent
            ge=expon(i,1)
            do j=2,dim
               if(expon(i,j).lt.ge) ge=expon(i,j)
            end do
c
c -- Add contribution to GCD
            if(ge.ne.0) GCD=GCD*(i**ge)
         end do
      end if
c
      return
      end
c
c
      subroutine INTEGERtoSTRING(i1,s1)
      implicit none
c
      integer maxP
      parameter(maxP=3)
c
      integer i,j,i1,i2,P(maxP)
      character*3 s1
      character*1 s(0:9),t(maxP)
c
      data s/'0','1','2','3','4','5','6','7','8','9'/
      data P/100,10,1/
c
c -- i1 is an integer used to generate a string
c -- s1 is the string to be returned of length 3(including leading zeroes)
      i2=i1
      do i=1,maxP
         j=i2/P(i)
         t(i)=s(j)
         i2=i2-j*P(i)
      end do
c
c -- integer as a string
      s1=t(1)//t(2)//t(3)
c
      return
      end
c
c
      subroutine List_Sort_Integer(list,nl,order)
      implicit none
c
c -- Adapted from Numerical Recipes, Press et al. page 227
      integer i,j,nl
      integer list(nl),order(nl),l1,o1
c
      do i=1,nl
         order(i)=i
      end do
c
c -- Sort list
      do i=2,nl
         l1=list(i)
         o1=order(i)
         do j=i-1,1,-1
            if(list(j).le.l1) goto 100 
            list(j+1)=list(j)
            order(j+1)=order(j)
         end do
         j=0
 100     continue
         list(j+1)=l1
         order(j+1)=o1
      end do
c
      return
      end
c
c
      subroutine List_Union_Integer(list,length)
      implicit none
c
      integer i,j,length,nU
      integer list(length),unique(length),order(length)
c
c -- Generate list of all unique elements in list
c
      nU=1
      unique(1)=list(1)
      do i=2,length
         do j=1,nU !i-1
            if(list(i).eq.unique(j)) goto 100
         end do
         nU=nU+1                ! Add element to list
         unique(nU)=list(i)
c
 100     continue               ! Element already in list
      end do
c
      call List_Sort_Integer(unique,nU,order)
c
      length=nU                 ! Overwrite original list and length
      do i=1,nU
         list(i)=unique(i)
      end do
c
      return
      end
c
c
      subroutine List_Min_Real(list,N,min)
      implicit none
c
      integer i,N
      real*8 list(N),min
c
      min=list(1)
      do i=2,N
         if(list(i).lt.min) min=list(i)
      end do
c
      return
      end
c
c
      subroutine DEBUG_OC(i,s,l)
      implicit none
c
      include 'param.inc'
c
      integer i,l,sl
      character*3 str1
      character*6 sClose
      character*26 sEnter,sExit
      character*50 s,FMT1
c
      sEnter='+++ Entering Subroutine ( '
      sExit='--- Exiting  Subroutine ( '
      sClose=' ) ***'
c
      call INTEGERtoSTRING(l,str1) ! Length of routine name
      sl=3
      if(l/10.ge.1) sl=2
      if(l/100.ge.1) sl=1
      FMT1='(a26,a'//str1(sl:3)//',a6)'
c
      if(i.eq.0) then
         write(LOGF,*)
         write(LOGF,FMT1) sEnter,s(1:l),sClose
      else
         write(LOGF,*)
         write(LOGF,FMT1) sExit,s(1:l),sClose
      end if
c
      return
      end
c
c
      subroutine Check_File(fileName,sL,LOGF)
      implicit none
c
      integer sL,LOGF
      character*50 fileName
      logical exitQ
c
      inquire(file=fileName(1:sL),EXIST=exitQ)
      if(.not.exitQ) then
         write(LOGF,*)
         write(LOGF,*) 'Warning: Check_File'
         write(LOGF,*) 'File does not exist: ',fileName(1:sL)
         stop
      end if
c
      return
      end
c
c
      real*8 function FUNC_NORM(r)
      implicit none
c
      real*8 r(3)
c
      FUNC_NORM=dsqrt(r(1)*r(1)+r(2)*r(2)+r(3)*r(3))
c
      return
      end
c
c
      real*8 function FUNC_DIST(r1,r2)
      implicit none
c
      real*8 r1(3),r2(3),r(3)
c
      r(1)=r1(1)-r2(1)
      r(2)=r1(2)-r2(2)
      r(3)=r1(3)-r2(3)
c
      FUNC_DIST=dsqrt(r(1)*r(1)+r(2)*r(2)+r(3)*r(3))
c
      return
      end
c
c
      subroutine EXIT_TERMINATE(i,LOGF)
      implicit none
c
      integer i,LOGF
c
      if(i.eq.1) then
         write(LOGF,*) 
         write(LOGF,*) 'The program will be terminated because some'
         write(LOGF,*) 'oxygens are not four coordinated.  Using the'
         write(LOGF,*) 'file Structure_Unit.txt and the various .xyz'
         write(LOGF,*) 'files, one should be able to construct a file'
         write(LOGF,*) 'named HBondsDef.txt in which all oxygens are'
         write(LOGF,*) 'four coordinated.'
      elseif(i.eq.2) then
         write(LOGF,*) 
         write(LOGF,*) 'The program will be terminated because',
     $        ' vertices &'
         write(LOGF,*) 'bonds did not form complete orbits or oxygen',
     $        ' atoms'
         write(LOGF,*) 'were not four coordinated.'
      else
         write(LOGF,*) 'Error: EXIT_TERMINATE'
         write(LOGF,*) 'Unknown option(i): ',i
         stop
      end if
c
      return
      end
c
c
      subroutine MemoryEstimate()
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer nI,nR,totalI,totalR
c
      call DEBUG_OC(0,'MemoryEstimate',14)
      write(LOGF,*)
      write(LOGF,*) 'Maximum array dimensions set as parameters:'
      write(LOGF,*) 'nGenMax:       ',nGenMax
      write(LOGF,*) 'nSymMax:       ',nSymMax
      write(LOGF,*) 'nVertUnitMax:  ',nVertUnitMax
      write(LOGF,*) 'nVertSimMax:   ',nVertSimMax
      write(LOGF,*) 'nBondsUnitMax: ',nBondsUnitMax
      write(LOGF,*) 'nBondsSimMax:  ',nBondsSimMax
      write(LOGF,*) 'nInvMax:       ',nInvMax
      write(LOGF,*) 'nBondPairsMax: ',nBondPairsMax
      write(LOGF,*)
      write(LOGF,*) 'Memory estimates for include files and',
     $     ' some routines:'
c
c -- Integers: 4 bytes
c -- Reals:    8 bytes(double precision)
c -- Logical:  4 bytes
c
c -- Initialize
      nI=0
      nR=0
c
      nI=nI+ 3                  ! maxPrime,jobHBONDS,jobGENINVAR
      nI=nI+ 2                  ! DEBUG,TERMINATE
      nR=nR+ 5                  ! TOL_ZERO,TOL_INTEGER,TOL_HBOND,TOL_VV,PI
c     
      call Write_Memory(nI,nR,'param.inc',9,LOGF)
c
      nI=0
      nR=0
c
c -- symmetry.inc
      nI=nI+ 3*4*nGenMax        ! Gen
      nI=nI+ 3*4*nSymMax        ! Sym
      nI=nI+ 5                  ! nGen,nGenMax,nSym,nSymMax,nSymUnit
c
      totalI=nI
      totalR=nR
c
c -- Write some output
      call Write_Memory(nI,nR,'symmetry.inc',12,LOGF)
c     
c -- structure.inc
      nI=0
      nR=0
c
      nI=nI+ 2*nBondsUnitMax    ! BondListUnit
      nI=nI+ 2*nBondsSimMax     ! BondListSim
      nI=nI+ 2*nBondsSimMax     ! BondListCov
c
      nI=nI+ 3*2*nBondsSimMax   ! HBonds
c
      nI=nI+ 3*nVertSimMax      ! Vertices
      nI=nI+ 3*nVertUnitMax     ! VertAsym
c
      nI=nI+ 13                 ! nVertAsym,nVertUnit,nVertSim,nVertCov,nVertUnitMax,
                                ! nVertSimMax,nBondsUnit,nBondsSim,nBondsCov,nBondsInv,
                                ! nBondsUnitMax,nBondsSimMax,LCD_VS,
c
      nR=nR+ 6                  ! LV_Param
      nR=nR+ 16*9               ! Cell,CellInv,Cell_Sim,CellInv_Sim,Cell_Cov,CellInv_Cov,
                                ! Cell_Inv,CellInv_Inv,LV_Unit,LVInv_Unit,LV_Sim,LVInv_Sim,
                                ! LV_Cov,LVInv_Cov,LV_Inv,LVInv_Inv
      nI=nI + 2                 ! ORTHOG,CELLBOUNDARY
c
      totalI=totalI+nI
      totalR=totalR+nR
c
      call Write_Memory(nI,nR,'structure.inc',13,LOGF)
c
c -- graphinvar.inc
      nI=0
      nR=0
c
c$$$      nI=nI+ nBondsSimMax       ! BondPerm_Mem_List
c$$$      nI=nI+ 2*nSymMax*nBondsSimMax ! BondPerm_Mem
c
      nI=nI+ 3*nSymMax*nInvMax  ! GraphInvar
      nI=nI+ nInvMax            ! GraphInvar_nTerms
      nI=nI+ nInvMax            ! GraphInvar_Type
      nI=nI+ nInvMax            ! GraphInvar_Norm
      nI=nI+ 2*nInvMax          ! GraphInvar_Pair
      nI=nI+ 3*2*2*nInvMax      ! GenBond
      nI=nI+ 2*nBondPairsMax*nInvMax ! BondPair
      nI=nI+ nInvMax            ! nBondPairs
c
      totalI=totalI+nI
      totalR=totalR+nR
c
      call Write_Memory(nI,nR,'graphinvar.inc',14,LOGF)
c
c -- InvarRoutines.f
      write(LOGF,*)
      write(LOGF,*) 'Invariant Routines:'
      nI=11+ 3*nSym
      nR=0
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Invar_Gen_B',11,LOGF)
c
      nI=15+ 3*nSym
      nR=0
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Invar_Gen_BB',12,LOGF)
c
      nI=15+ 3*nSym
      nR=0
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Invar_Gen_BC',12,LOGF)
c
      nI=4+ 3*nSym
      nR=0
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Invar_Contract',14,LOGF)
c
      nI=5+ 3*nSym+ 2*nBondsInv+ 3*nBondsInv*nBondsInv
      nR=0
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Invar_Sort',10,LOGF)
c
c -- ReadWrite.f
      write(LOGF,*)
      write(LOGF,*) 'ReadWrite Routines:'
      nI=13+ nGenMax*3+ 3*5*nGenMax
      nR=0
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Read_Input',10,LOGF)
c
      nI=6+ 100
      nR=100+ 3
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Read_Line_Single',16,LOGF)
c
      nI=8+ 100
      nR=100+ 3*2
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Read_Line_Multiple',18,LOGF)
c
      nI=8+ nBondsUnitMax*6+ 3*2*2*nBondsUnitMax
      nR=0
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Read_Bonds',10,LOGF)
c
      nI=8+ nVertUnitMax*3+ 3*2*nVertUnitMax
      nR=0
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Read_Asym',9,LOGF)
c
      nI=10
      nR=3+ 3*8+ 3*nVertSimMax+ 3*nBondsSimMax
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Write_XYZ',9,LOGF)
c
c -- Structure.f
      write(LOGF,*)
      write(LOGF,*) 'Structure Routines:'
      nI=12+ nVertUnitMax
      nR=2+2*3+3*nVertUnitMax+3*27
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Generate_Vertices_HBonds',24,LOGF)
c
      nI=8+ nVertUnitMax+ 3*27
      nR=5+ 3*nVertUnitMax+ 2*27
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Generate_HBonds_Asym',20,LOGF)
c
      nI=4
      nR=2+2*3+3*nVertUnitMax
      totalI=totalI+nI
      totalR=totalR+nR
      call Write_Memory(nI,nR,'Generate_Vertices_Asym',22,LOGF)
c
      write(LOGF,*)
      call Write_Memory(totalI,totalR,'Total',5,LOGF)

      call DEBUG_OC(1,'MemoryEstimate',14)
c
      return
      end
