!#########################################################################
!  Graph Invariant Routines for MkInvar: 
!     GenInvariant   -- Generate invariants for cell of interest.
!     GenInvar_ALL   -- Generate all unique invariants up to and
!                        including 2nd order b[j]b[k] & b[j]c[k]
!     GenInvar_BP    -- Only generate those invariants for which 
!                        a generating bond pair is supplied.
!     GenInvar_Cover -- Use covering cell invariants to generate 
!                        simulation cell invariants.
!
!#########################################################################    
      subroutine GenInvariant()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
      include 'graphinvar.inc'
c
      integer i,j
c
      if(DEBUG) call DEBUG_OC(0,'GenInvariant',12)
c
c -- Determine which cell we'll calculate invariants for and make a
c -- copy of the corresponding lattice vectors.
      if(COVERCELL) then
         if(DEBUG) write(LOGF,*) 'Copying info from cover cell'
         nBondsInv=nBondsCov
         do i=1,3
            do j=1,3
               Cell_Inv(i,j)=Cell_Cov(i,j)
               CellInv_Inv(i,j)=CellInv_Cov(i,j)
               LV_Inv(i,j)=LV_Cov(i,j)
               LVInv_Inv(i,j)=LVInv_Cov(i,j)
            end do
         end do
      elseif(SIMCELL) then
         if(DEBUG) write(LOGF,*) 'Copying info from simulation cell'
         nBondsInv=nBondsSim
         do i=1,3
            do j=1,3
               Cell_Inv(i,j)=Cell_Sim(i,j)
               CellInv_Inv(i,j)=CellInv_Sim(i,j)
               LV_Inv(i,j)=LV_Sim(i,j)
               LVInv_Inv(i,j)=LVInv_Sim(i,j)
            end do
         end do
      else
         if(DEBUG) write(LOGF,*) 'Copying info from unit cell'
         nBondsInv=nBondsUnit
         do i=1,3
            do j=1,3
               Cell_Inv(i,j)=Cell(i,j)
               CellInv_Inv(i,j)=CellInv(i,j)
               LV_Inv(i,j)=LV_Unit(i,j)
               LVInv_Inv(i,j)=LVInv_Unit(i,j)
            end do
         end do
      end if
c
c -- Initialization
c$$$      do i=1,nBondsInv
c$$$         BondPerm_Mem_List(i)=0
c$$$      end do
c
c -- Generate invariants
      if(jobGENINVAR.eq.1) then ! Generate all invariants
         call GenInvar_ALL()
      elseif(jobGENINVAR.eq.2) then ! Generate small set of invariants
         call GenInvar_BP()
      end if
c
c -- Write invariants to file
      if(DEBUG) then 
         write(LOGF,*)
         write(LOGF,*) 'Graph Invariants written to GraphInvar.txt.'
      end if
      call Write_GraphInvar('GraphInvar.txt',14)
c
c -- Write generating bond pairs to file if invariants were not
c -- generated from a list of bond pairs.
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) 'Writing generating bond pair info to file.'
      end if
      call Write_GenBondPair('GenBondPair.txt',15,'BondPairs.txt',13)
c
      if(DEBUG) call DEBUG_OC(1,'GenInvariant',12)
c
      return
      end
c
c
      subroutine GenInvariantCover()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
      include 'graphinvar.inc'
c
      if(DEBUG) call DEBUG_OC(0,'GenInvariantCover',17)
c
c -- Take the current list of invariants, apply bond constraints, 
c -- and generate a list of invariants for the simulation cell.
      call GenInvarC()
c
c -- Write invariants to file
      if(DEBUG) then 
         write(LOGF,*)
         write(LOGF,*) 'Graph Invariants written to GraphInvarC.txt.'
      end if
      call Write_GraphInvar('GraphInvarC.txt',15)
c
c -- Write generating bond pairs to file if invariants were not
c -- generated from a list of bond pairs.
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) 'Writing generating bond pair info to'
     $        ,' BondPairsC.txt.'
      end if
      call Write_GenBondPair('GenBondPairC.txt',16,'BondPairsC.txt',14)
c
      if(DEBUG) call DEBUG_OC(1,'GenInvariantCover',17)
c
      return
      end
c
c
      subroutine GenInvar_ALL()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
      include 'graphinvar.inc'
c
      if(DEBUG) call DEBUG_OC(0,'GenInvar_ALL',12)
c
c -- Initialization
      nInv=0
c
c -- Generate all unique 1st order invariants: b[j]
      call Invar_Gen_ALL_B()
c$$$c
c$$$c -- Method 1
c$$$c
c$$$c -- Generate all unique 2nd order invariants: b[j]b[k]
c$$$      call Invar_Gen_ALL_BB()
c$$$c
c$$$c -- Generate all unique 2nd order invariants: b[j]c[k]
c$$$      call Invar_Gen_ALL_BC()
c
c -- Method 2: Noticeable speed up for large systems.
c
c -- Generate all unique 2nd order invariants: b[j]c[k]
      call Invar_Gen_ALL_BC()
c
c -- Generate all unique 2nd order invariants: b[j]b[k]
      call Invar_Gen_ALL_BB_from_BC()
c
      if(DEBUG) call DEBUG_OC(1,'GenInvar_ALL',12)
c
      return
      end
c
c
      subroutine GenInvar_BP()
      implicit none
c
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
      include 'graphinvar.inc'
c
      integer i,k,nTerms,IB(2),JB(2),n(2)
c
      if(DEBUG) call DEBUG_OC(0,'GenInvar_BP',11)
c
c -- Enough memory allocated?
      if(nInv.ge.nInvMax) then
         write(LOGF,*) 'Error: GenInvar_BP'
         write(LOGF,*) 'Not enough memory allocated'
         write(LOGF,*) '# of invariants found(nInv): ',nInv
         write(LOGF,*) 'Maximum allowed(nInvMax):    ',nInvMax
         stop
      end if
c
      if(DEBUG) then 
         n(1)=0
         n(2)=0
         do i=1,nInv
            if(GraphInvar_Type(i).eq.0) n(1)=n(1)+1
            if(GraphInvar_Type(i).eq.1) n(2)=n(2)+1
         end do
         write(LOGF,*)
         write(LOGF,*) 'Generating Bond Pairs:'
         write(LOGF,*) '# of 1st order invariants, b[j]:     ',n(1)
         write(LOGF,*) '# of 2nd order invariants, b[k]b[k]: ',n(2)
         write(LOGF,*) '# of 2nd order invariants, b[k]c[k]: ',
     $        nInv-n(1)-n(2)
      end if
c
c -- Identify bond index for generating bond pairs.  If this were to fail
c -- then the invariants generated would be garbage.  
      do i=1,nInv
         call FindBond(GenBond(1,1,1,i),nBondsInv,IB,HBonds)
         GraphInvar_Pair(1,i)=IB(2)
         if(GraphInvar_Type(i).gt.0) then ! 2nd order graph invariant
         call FindBond(GenBond(1,1,2,i),nBondsInv,JB,HBonds)
            GraphInvar_Pair(2,i)=JB(2)
         else
            GraphInvar_Pair(2,i)=0
         end if
c
         if((IB(2).eq.0).or.
     $        ((JB(2).eq.0).and.(GraphInvar_Type(i).gt.0))) then
            write(LOGF,*)
            write(LOGF,*) 'Error: GenInvar_BP()'
            write(LOGF,*) 'Generating Bond Pair not found: ',i
            write(LOGF,*) 'Invariant Type: ',GraphInvar_Type(i)
            write(LOGF,*) 'Bond 1: ',(GenBond(k,1,1,i),k=1,3),'   ',
     $           (GenBond(k,2,1,i),k=1,3)
            write(LOGF,*) 'IB: ',IB(1),'  ',IB(2)
            if(GraphInvar_Type(i).gt.0) then
               write(LOGF,*) 'Bond 2: ',(GenBond(k,1,2,i),k=1,3),'   ',
     $           (GenBond(k,2,2,i),k=1,3)
               write(LOGF,*) 'JB: ',JB(1),'  ',JB(2)
            end if
            stop
         end if
      end do
c
c -- Loop over generating bond pairs and generate corresponding invariants.
      do i=1,nInv
         nTerms=nSym
         nBondPairs(i)=1
         BondPair(1,1,i)=GraphInvar_Pair(1,i)
         BondPair(2,1,i)=GraphInvar_Pair(2,i)
         if(GraphInvar_Type(i).eq.0) then
            call Invar_Gen_B(GraphInvar_Pair(1,i),GraphInvar(1,1,i))
         elseif(GraphInvar_Type(i).eq.1) then
            call Invar_Gen_BB(GraphInvar_Pair(1,i),GraphInvar_Pair(2,i),
     $           GraphInvar(1,1,i))
         elseif(GraphInvar_Type(i).eq.2) then
            call Invar_Gen_BC(GraphInvar_Pair(1,i),GraphInvar_Pair(2,i),
     $           GraphInvar(1,1,i))
         else
            write(LOGF,*) 'Error: GenInvar_BP'
            write(LOGF,*) 'Unknown Invariant Type: ',GraphInvar_Type(i)
            stop
         end if
c
         call Invar_Contract(GraphInvar(1,1,i),nTerms)
         call Invar_Sort(GraphInvar(1,1,i),nTerms)
         GraphInvar_nTerms(i)=nTerms
         GraphInvar_Norm(i)=nSym
      end do
c
      if(DEBUG) then 
         n(1)=0
         n(2)=0
         do i=1,nInv
            if(GraphInvar_Type(i).eq.0) n(1)=n(1)+1
            if(GraphInvar_Type(i).eq.1) n(2)=n(2)+1
         end do
         write(LOGF,*)
         write(LOGF,*) 'Invariants Generated:'
         write(LOGF,*) '# of 1st order invariants, b[j]:     ',n(1)
         write(LOGF,*) '# of 2nd order invariants, b[k]b[k]: ',n(2)
         write(LOGF,*) '# of 2nd order invariants, b[k]c[k]: ',
     $        nInv-n(1)-n(2)
         call DEBUG_OC(1,'GenInvar_BP',11)
      end if
c
      return
      end
c
c
      subroutine GenInvarC()
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer i,j,k,IG,iInv,iTerm,iNew,b1,b2,n(2)
      logical test
c
      if(DEBUG) call DEBUG_OC(0,'GenInvar_Sim_from_Cover',23)
c
c -- Loop over invariants replacing covering cell bond variables with 
c -- simulation cell bond variables.  Unique invariants are placed at 
c -- beginning of list.
      iNew=0
      do iInv=1,nInv
         do iTerm=1,GraphInvar_nTerms(iInv)
            if(GraphInvar_Type(iInv).eq.0) then ! 1st order invariants
               GraphInvar(1,iTerm,iNew+1)=GraphInvar(1,iTerm,iInv)
               b1=GraphInvar(2,iTerm,iInv)
               GraphInvar(2,iTerm,iNew+1)=BondMap(b1)
c
            elseif(GraphInvar_Type(iInv).gt.0) then ! 2nd order invariants
               b1=BondMap(GraphInvar(2,iTerm,iInv)) ! 1st bond
               b2=BondMap(GraphInvar(3,iTerm,iInv)) ! 2nd bond
c
               GraphInvar(1,iTerm,iNew+1)=GraphInvar(1,iTerm,iInv)
               GraphInvar(2,iTerm,iNew+1)=b1
               GraphInvar(3,iTerm,iNew+1)=b2
               if((GraphInvar_Type(iInv).eq.1).and.(b1.gt.b2)) then ! Order terms
                  GraphInvar(2,iTerm,iNew+1)=b2
                  GraphInvar(3,iTerm,iNew+1)=b1
               end if
            end if
         end do
c
         GraphInvar_Type(iNew+1)=GraphInvar_Type(iInv)
         GraphInvar_nTerms(iNew+1)=GraphInvar_nTerms(iInv)
         GraphInvar_Norm(iNew+1)=GraphInvar_Norm(iInv) 
c
         call Invar_Contract(GraphInvar(1,1,iNew+1),
     $        GraphInvar_nTerms(iNew+1))
c
         call Invar_Sort(GraphInvar(1,1,iNew+1),
     $        GraphInvar_nTerms(iNew+1))
c
c -- Check if invariant is new
         if(iNew.eq.0) then
            IG=0
         else
            call Invar_Check(iNew+1,IG)
         end if
c
c -- Update invariant descriptors
         if(IG.eq.0) then
            iNew=iNew+1
c
            GraphInvar_Pair(1,iNew)=BondMap(GraphInvar_Pair(1,iInv))
            GraphInvar_Pair(2,iNew)=BondMap(GraphInvar_Pair(2,iInv))
c
c -- Generating bond pairs
            nBondPairs(iNew)=nBondPairs(iInv)
            j=0
            do i=1,nBondPairs(iNew)
               test=.false.
               b1=BondMap(BondPair(1,i,iInv))
               b2=BondMap(BondPair(2,i,iInv))
c
c -- Order the bond variables if b[j]b[k] type invariant
               if((GraphInvar_Type(iNew).eq.1).and.(b2.lt.b1)) then
                  BondPair(1,j+1,iNew)=b2
                  BondPair(2,j+1,iNew)=b1
               else
                  BondPair(1,j+1,iNew)=b1
                  BondPair(2,j+1,iNew)=b2
               end if
c
c -- Remove duplicate bond pairs
               if(j.eq.0) then
                  test=.true.
               else
                  do k=1,j
                     if((BondPair(1,k,iNew).eq.BondPair(1,j+1,iNew))
     $                    .and.(BondPair(2,k,iNew).eq.
     $                    BondPair(2,j+1,iNew))) goto 100
                  end do
                  test=.true.
               end if
               if(test) j=j+1
 100           continue         ! Bond pairs are the same
               nBondPairs(iNew)=j
            end do
         end if
      end do
      nInv=iNew
c
      if(DEBUG) then 
         n(1)=0
         n(2)=0
         do i=1,nInv
            if(GraphInvar_Type(i).eq.0) n(1)=n(1)+1
            if(GraphInvar_Type(i).eq.1) n(2)=n(2)+1
         end do
         write(LOGF,*)
         write(LOGF,*) '# of 1st order invariants, b[j]:     ',n(1)
         write(LOGF,*) '# of 2nd order invariants, b[k]b[k]: ',n(2)
         write(LOGF,*) '# of 2nd order invariants, b[k]c[k]: ',
     $        nInv-n(1)-n(2)
         write(LOGF,*)
         write(LOGF,*) '# of invariants in simulation cell:  ',nInv
         write(LOGF,*) 'These invariants may have algebraic linear',
     $        ' dependencies.'
         call DEBUG_OC(1,'GenInvar_Sim_from_Cover',23)
      end if
c
      return
      end
