!#########################################################################
!  Read & Write Routines for MkInvar: 
!     GetFileName              -- Retrieve filename from command line.
!     Read_Input               -- Read cell info and symmetry elements
!                                  from file.
!     Read_Line_Single         -- Read line of input file and return a
!                                  single vector of dimension 3 formatted
!                                  as real*8.
!     Write_Group              -- Write symmetry elements of group to file
!     Read_Symmetry            -- Read symmetry element from file
!     Write_Line_Symmetry      -- Write formatted symmetry element to file.
!     Read_Group               -- Read space group info and generators
!     Get_SpaceGroup_FileName  -- Deterimine full path to space group file
!     Read_Bonds               -- Read H-bonds from file
!     Read_Asym                -- Read asymmetric unit from file.
!     Write_XYZ                -- Write atomic coordinates to xyz file.
!     Write_Structure          -- Write lattice vectors, vertices, and HBonds
!                                  to file.
!     Write_GraphInvar         -- Write graph invariants to file.
!     Write_GenBondPair        -- Write generating bond pairs to file.
!     Read_BondPair            -- Read generating bond pairs from file
!     Write_Enumeration_Input  -- Write input for enumeration program
!
!#########################################################################   
      subroutine Get_FileName(f1,length,LOGF)
      implicit none
c
      integer i,length,iargc,LOGF
      character*50 f1
c
c -- There should only be one entry on the command line
      i=iargc()
      if(i.ne.1) then
         write(LOGF,*) 'Error: GetFileName'
         write(LOGF,*) 'There is an incorrect number of entries on',
     $        ' the command line.'
         write(LOGF,*) 'Expecting: ',1
         write(LOGF,*) 'Found:     ',i
         stop
      end if
c
      call getarg(1,f1)
c
c -- Determine length of string.
      do i=1,50
         if(f1(i:i).eq.' ') goto 100
         length=i
      end do
      write(LOGF,*) 'Error: GetFileName'
      write(LOGF,*) 'No empty spaces were found'
      write(LOGF,*) 'May be necessary to increase dimension of',
     $     ' filename: ',50
      stop
 100  continue
c
      return
      end
c
c
      subroutine Read_Input()
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
c
      integer i,i1,iUnit,iUnit2,j,k,l,denom(nGenMax*3),FileNameLength,
     $     Gen_Temp(3,5,nGenMax),LCD_GEN,SpaceGroup(2)
      character*3 CellType
      character*50 InputFile,SpaceGroupFile
c
      if(DEBUG) call DEBUG_OC(0,'ReadInput',9)
c
c -- Retrieve input filename from command line
c$$$      call Get_FileName(InputFile,FileNameLength,LOGF)
      InputFile='MkInvar.inp'
      FileNameLength=11
c
      iUnit=1
      open(unit=iUnit,file=InputFile(1:FileNameLength))
c$$$c
c$$$c -- Verbose Output?
c$$$      DEBUG=.false.
c$$$      read(iUnit,*) i
c$$$      if(i.eq.1) DEBUG=.true.
c
c -- Read simulation cell dimensions
      read(iUnit,*) (SimCellDim(j),j=1,3)
      if(DEBUG) write(LOGF,111) (SimCellDim(j),j=1,3)
c
      if((SimCellDim(1).eq.SimCellDim(2)).and.
     $     (SimCellDim(2).eq.SimCellDim(3))) then
         if(SimCellDim(1).gt.1) then
            SIMCELL=.true.
         else
            write(LOGF,*) ' -- This is a unit cell.'
         end if
      else
c
c -- Simulation cell larger than unit cell w/o symmetry of unit cell
         SIMCELL=.true.
         COVERCELL=.true.
         call Lowest_Common_Multiple(SimCellDim,3,CovCellDim)
         if(DEBUG) then 
            write(LOGF,*)
            write(LOGF,*) 'A covering cell used for this calculation.'
            write(LOGF,112) (CovCellDim,j=1,3)
         end if
      end if
c
c -- Lattice vectors stored as matrix of row vectors
c -- The matrix is read is if they were column vectors
c -- We revert back to row vectors by taking the transpose.
      call Read_Line_Single(iUnit,Cell(1,1),LOGF)
      call Read_Line_Single(iUnit,Cell(1,2),LOGF)
      call Read_Line_Single(iUnit,Cell(1,3),LOGF)
      call Matrix_Transpose(Cell)
      if(DEBUG) then
         write(LOGF,*) 
         write(LOGF,*) 'Matrix of Cell Vectors:'
         write(LOGF,11) (Cell(1,j),j=1,3)
         write(LOGF,11) (Cell(2,j),j=1,3)
         write(LOGF,11) (Cell(3,j),j=1,3)
      end if
c
c -- If the matrix of cell vectors is the identity matrix, this cell
c -- is considered to be a primitive unit cell.  Otherwise, these cell 
c -- vectors form a non-primitive unit cell.  This has the consequence
c -- of requiring the translation group for a larger covering cell to 
c -- correctly generate the structure and symmetry elements of the group.
c -- This covering cell is not necessarily related to the covering cell
c -- needed to generate invariants for an N1xN2xN3 simulation cell where
c -- N1, N2, and N3 are different.
      PRIMITIVE=.false.
      do i=1,3
         do j=1,3
            if(i.eq.j) then
               if(abs(Cell(i,j)-1.d0).gt.TOL_ZERO) goto 100
            else
               if(abs(Cell(i,j)).gt.TOL_ZERO) goto 100
            end if
         end do
      end do
      PRIMITIVE=.true.
 100  continue
c
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) 'Primitive Cell? ',PRIMITIVE
      end if
c
c -- Identify range of indices for covering cell if non-primitive
      if(.not.PRIMITIVE) then
         do i=1,3               ! Loop over columns
            NonPrimRange(1,i)=0 ! Min
            NonPrimRange(2,i)=0 ! Max
            do j=1,3            ! Loop over rows
               k=nint(Cell(j,i)-1.d0)
               if(k.lt.NonPrimRange(1,i)) NonPrimRange(1,i)=k
c
               k=nint(Cell(j,i)+1.d0)
               if(k.gt.NonPrimRange(2,i)) NonPrimRange(2,i)=k
            end do
         end do
c
         if(DEBUG) then
            write(LOGF,*) 'Range of indices for covering cell: '
            write(LOGF,*) 'Cell Vector 1: ',
     $           (NonPrimRange(i,1),' ',i=1,2)
            write(LOGF,*) 'Cell Vector 2: ',
     $           (NonPrimRange(i,2),' ',i=1,2)
            write(LOGF,*) 'Cell Vector 3: ',
     $           (NonPrimRange(i,3),' ',i=1,2)
         end if
      end if
c
c -- Calculate inverse of matrix of lattice vectors
      call Matrix_Inverse(Cell,CellInv,LOGF)
      if(DEBUG) then
         write(LOGF,*) 
         write(LOGF,*) 'Inverse of Matrix of Cell Vectors:'
         write(LOGF,11) (CellInv(1,j),j=1,3)
         write(LOGF,11) (CellInv(2,j),j=1,3)
         write(LOGF,11) (CellInv(3,j),j=1,3)
      end if
c
c -- Read generators of symmetry group
      read(iUnit,*) SpaceGroup(1),SpaceGroup(2)
      if(DEBUG) then 
         write(LOGF,*)
         write(LOGF,*) 'SpaceGroup Index:      ',SpaceGroup(1)
         write(LOGF,*) 'SpaceGroup Convention: ',SpaceGroup(2)
      end if
      call Get_SpaceGroup_FileName(iUnit,SpaceGroup,SpaceGroupFile,l,
     $     DEBUG,LOGF)
c
c -- Does file exist?
      call Check_File(SpaceGroupFile,l,LOGF)
      iUnit2=2
      open(iUnit2,file=SpaceGroupFile(1:l))
      read(iUnit2,*) i
      if(i.ne.SpaceGroup(1)) then
         write(LOGF,*) 'Error: ReadInput'
         write(LOGF,*) 'Space Group index does not agree w/ file'
         write(LOGF,*) 'Index from database file : ',i
         write(LOGF,*) 'Index from input file:     ',SpaceGroup(1)
         stop
      end if
c     
      if(DEBUG) then 
         write(LOGF,*) 
         write(LOGF,*) 'Reading space group file...'
      end if
      call Read_Group(iUnit2,nGen,Gen_Temp,nGenMax,nSymUnit,CellType)
      if(DEBUG) write(LOGF,'(28x,a10)') 'Successful'
      close(iUnit2)
c
c -- Determine lowest common multiple for the denominator of components of
c -- translations in the space group
      l=0
      do i=1,nGen               ! Generate list of denominators
         do j=1,3
            l=l+1
            denom(l)=Gen_Temp(j,5,i)
         end do
      end do
c
      call List_Union_Integer(denom,l) ! Union of the list of denominators
      call Lowest_Common_Multiple(denom,l,LCD_GEN)
c
      if(DEBUG) then 
         write(LOGF,*) 
         write(LOGF,121) (denom(i),i=1,l)
         write(LOGF,*) 'Lowest Common Denominator of Trans. Vectors: ',
     $        LCD_GEN
      end if
c
c -- Read lattice parameters which are used to construct lattice
c -- vectors for the unit cell of the crystal.  These vectors form
c -- a cartesian basis for the crystal.
      read(iUnit,*) (LV_Param(i),i=1,6)
      call Check_LatticeParam(CellType)
      call Generate_LatticeVector()
      call Matrix_Inverse(LV_Unit,LVInv_Unit,LOGF)
c
c -- If we are working with a non-primitive cell, transform the unit cell vectors
c -- to non-primitive vectors.
      if(.not.PRIMITIVE) then
         call Matrix_Multiply_MN(Cell,LV_Unit,LV_Unit_NonPrim)
         call Matrix_Inverse(LV_Unit_NonPrim,LVInv_Unit_NonPrim,LOGF)
      end if
c
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) 'Matrix of Unit Cell Lattice Vectors:'
         write(LOGF,12) (LV_Unit(1,i),i=1,3)
         write(LOGF,12) (LV_Unit(2,i),i=1,3)
         write(LOGF,12) (LV_Unit(3,i),i=1,3)
         write(LOGF,*)
         write(LOGF,*) 'Inverse of Unit Cell Lattice Vectors:'
         write(LOGF,12) (LVInv_Unit(1,i),i=1,3)
         write(LOGF,12) (LVInv_Unit(2,i),i=1,3)
         write(LOGF,12) (LVInv_Unit(3,i),i=1,3)
         if(.not.PRIMITIVE) then
            write(LOGF,*)
            write(LOGF,*) 'Matrix of Non-Primitive Lattice Vectors:'
            write(LOGF,12) (LV_Unit_NonPrim(1,i),i=1,3)
            write(LOGF,12) (LV_Unit_NonPrim(2,i),i=1,3)
            write(LOGF,12) (LV_Unit_NonPrim(3,i),i=1,3)
            write(LOGF,*)
            write(LOGF,*) 'Inverse of Non-Primitive Lattice Vectors:'
            write(LOGF,12) (LVInv_Unit_NonPrim(1,i),i=1,3)
            write(LOGF,12) (LVInv_Unit_NonPrim(2,i),i=1,3)
            write(LOGF,12) (LVInv_Unit_NonPrim(3,i),i=1,3)
         end if
      end if
c
c -- Minimum allowed OO distance
      read(iUnit,*) TOL_VV
c
c -- Minimum & Maximum allowed OO distance for H-Bonds.
      read(iUnit,*) TOL_HBOND(1),TOL_HBOND(2)
c
      if(DEBUG) then 
         write(LOGF,*)
         write(LOGF,131) TOL_VV
         write(LOGF,132) TOL_HBOND(1)
         write(LOGF,133) TOL_HBOND(2)
      end if
c
c -- How do we determine H-Bonds?
      read(iUnit,*) jobHBONDS
      if(DEBUG) then
         write(LOGF,*) 
         write(LOGF,141) jobHBonds
         write(LOGF,*) '0: Read vertices from VertDef.txt.'
         write(LOGF,*) '1: Read H-bonds from HBondsDef.txt.'
      end if
c
c -- How do we generate invariants?
      read(iUnit,*) jobGENINVAR
      if(DEBUG) then
         write(LOGF,*) 
         write(LOGF,142) jobGENINVAR
         write(LOGF,*) '0: Do not generate invariants'
         write(LOGF,*) '1: Generate all invariants.'
         write(LOGF,*) '2: Read bond pairs from BondPairDef.txt.'
      end if
c
      if((jobGENINVAR.lt.0).or.(jobGENINVAR.gt.2)) then
         write(LOGF,*) 
         write(LOGF,*) 'Error: Read_Input'
         write(LOGF,*) 'Invalid option for invariants(jobGENINVAR): ',
     $        jobGENINVAR
         write(LOGF,*) 'Expecting one of the following'
         write(LOGF,*) '0: Do not generate invariants'
         write(LOGF,*) '1: Generate all invariants.'
         write(LOGF,*) '2: Read bond pairs from BondPairDef.txt.'
         stop
      end if
c
c -- Is this the covering cell for a non-primitive cell that does not
c -- have the same symmetry as the crystal spacegroup?
      read(iUnit,*) i
      if(i.eq.0) then
         COVERNP=.true.
         if(DEBUG) then
            write(LOGF,*) 
            write(LOGF,*) '********************************************'
            write(LOGF,*) ' Use first cell as covering cell for the',
     $           ' cell that follows.'
            write(LOGF,*) '********************************************'
         end if
      else
         COVERNP=.false.
      end if
c
c -- Read origin shift and non-primitive lattice vectors is necessary
      if(COVERNP) then
         call Read_Line_Single(iUnit,CellC(1,1),LOGF)
         call Read_Line_Single(iUnit,CellC(1,2),LOGF)
         call Read_Line_Single(iUnit,CellC(1,3),LOGF)
         call Matrix_Transpose(CellC)
         call Matrix_Inverse(CellC,CellCInv,LOGF)
c
         if(DEBUG) then
            write(LOGF,*) 
            write(LOGF,*) 'Matrix of Cell Vectors:'
            write(LOGF,11) (CellC(1,j),j=1,3)
            write(LOGF,11) (CellC(2,j),j=1,3)
            write(LOGF,11) (CellC(3,j),j=1,3)
            write(LOGF,*) 
            write(LOGF,*) 'Inverse of Matrix of Cell Vectors:'
            write(LOGF,11) (CellCInv(1,j),j=1,3)
            write(LOGF,11) (CellCInv(2,j),j=1,3)
            write(LOGF,11) (CellCInv(3,j),j=1,3)
         end if
c
         call Matrix_Multiply_MN(CellC,LV_Unit,LV_Unit_NonPrim)
         call Matrix_Inverse(LV_Unit_NonPrim,LVInv_Unit_NonPrim,LOGF)
c
         if(DEBUG) then
            write(LOGF,*)
            write(LOGF,*) 'Matrix of Non-Primitive Lattice Vectors:'
            write(LOGF,12) (LV_Unit_NonPrim(1,i),i=1,3)
            write(LOGF,12) (LV_Unit_NonPrim(2,i),i=1,3)
            write(LOGF,12) (LV_Unit_NonPrim(3,i),i=1,3)
            write(LOGF,*)
            write(LOGF,*) 'Inverse of Non-Primitive Lattice Vectors:'
            write(LOGF,12) (LVInv_Unit_NonPrim(1,i),i=1,3)
            write(LOGF,12) (LVInv_Unit_NonPrim(2,i),i=1,3)
            write(LOGF,12) (LVInv_Unit_NonPrim(3,i),i=1,3)
         end if
      end if
c
c -- Check that the first cell is an appropriate covering cell
      if(COVERNP) then
c
         write(LOGF,*)
         write(LOGF,*) '**************************************'
         write(LOGF,*) ' Requires routine to confirm that the '
         write(LOGF,*) ' non-primitive cell is completely'
         write(LOGF,*) ' encompassed by the covering cell.'
         write(LOGF,*) '**************************************'
c
         if(COVERCELL) then 
            write(LOGF,*)
            write(LOGF,*) 'Coeffcients for covering cell are not',
     $           ' identical.'
            write(LOGF,*) 'SimCellDim: ',(SimCellDim(j),j=1,3)
            stop
         end if
c
         if(.not.SIMCELL) then
            write(LOGF,*)
            write(LOGF,*) 'Simulation cell must be larger than',
     $           ' primitive unit cell.'
            write(LOGF,*) 
            write(LOGF,*) 'In current version of code, simulation',
     $           ' cell dimensions must be larger than one.'
         end if
c
         if(.not.PRIMITIVE) then
            write(LOGF,*)
            write(LOGF,*) 'Covering cell is non-primitive'
            stop
         end if
      end if
c
c -- Close input file
      close(iUnit)
c
c -- Read structure from file
      if(jobHBonds.eq.0) then   ! Read asymetric unit from file
         call Read_Asym('VertDef.txt',11,LCD_GEN)
      elseif(jobHBonds.eq.1) then ! Read H-Bonds from file
         call Read_Bonds('HBondsDef.txt',13,LCD_GEN)
      else                      ! Invalid Option
         write(LOGF,*) 
         write(LOGF,*) 'Error: Read_Input'
         write(LOGF,*) 'Invalid option for determing H-bonds',
     $        '(jobHBonds): ',jobHBonds
         stop
      end if
c
c -- Scale numerator of translation vector by LCD_VS to treat them as integer.
      do i=1,nGen               ! Loop over generators
         do j=1,4
            do k=1,3
               if(j.eq.4) then
                  i1=LCD_VS/Gen_Temp(k,5,i)
                  Gen(k,4,i)=Gen_Temp(k,4,i)*i1 ! Translation vector
               else
                  Gen(k,j,i)=Gen_Temp(k,j,i) ! Rotation matrix
               end if
            end do
         end do
      end do
c
      if(DEBUG) then
         write(LOGF,*) 
         write(LOGF,*) 'Generators of symmetry group:'
         write(LOGF,*) '# of generators: ',nGen
         do i=1,nGen
            call Write_Line_Symmetry(0,Gen(1,1,i),LCD_VS,LOGF)
         end do
      end if
c
c -- Scale matrix of cell vectors by LCD_VS
      do i=1,3
         do j=1,3
            Cell(i,j)=Cell(i,j)*dfloat(LCD_VS)
            CellInv(i,j)=CellInv(i,j)/dfloat(LCD_VS)
            if(COVERNP) then
               CellC(i,j)=CellC(i,j)*dfloat(LCD_VS)
               CellCInv(i,j)=CellCInv(i,j)/dfloat(LCD_VS)
            end if
         end do
      end do
c
c -- Read structure from file
      if(jobGENINVAR.eq.2) call Read_BondPair('BondPairsDef.txt',16)
c
      if(DEBUG) call DEBUG_OC(1,'ReadInput',9)
c
 11   format(3(1x,f5.2))
 12   format(3(1x,f10.5))
c
 111  format(1x,'Simulation Cell:',12x,3(i3,1x))
 112  format(1x,'Covering Cell Dimensions:',3x,3(i3,1x))
c
 121  format(1x,'Unique denominators of Trans. Vectors:',3(1x,i4))
c
 131  format(1x,'Minimum allowed OO distance:',10x,f12.6)
 132  format(1x,'Minimum allowed H-bonded OO distance:',1x,f12.6)
 133  format(1x,'Maximum allowed H-bonded OO distance:',1x,f12.6)
c
 141  format(1x,'Method to determine H-bonds(jobHBONDS):',7x,i10)
 142  format(1x,'Method to generate invariants(jobGENINVAR):',3x,i10)
c
      return
      end
c
c
      subroutine Read_Line_Single(iUnit,array,LOGF)
      implicit none
c
      integer nEntries
      parameter(nEntries=6)
c
      integer i,io,iUnit,j,inputI(nEntries),LOGF
      real*8 inputR(nEntries),array(nEntries/2)
c
c -- Read line first as array of integers
      read(iUnit,*) (inputI(i),i=1,nEntries)
c
c -- Reread line as array of real*8
      backspace (1,IOSTAT=io,err=10)
      read(iUnit,*) (inputR(i),i=1,nEntries)
c
c -- Each entry is the ratio of two integers.  If the numerator is 
c -- zero, then the entry is taken to be the denominator as a real*8.
      do i=1,nEntries/2
         j=(i-1)*2+1            ! index of ith numerator
         if(inputI(j).eq.0) then
            array(i)=inputR(j+1)
         else
            array(i)=dfloat(inputI(j))/dfloat(inputI(j+1))
         end if
      end do
c
      return
c
 10   write(LOGF,*) 'Error: ReadLine_Single'
      write(LOGF,*) 'Backspace error'
      write(LOGF,*) 'IOSTAT: ',io
      stop
c
      end
c
c
      subroutine Write_Group(iUnit,LCD)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
c
      integer i,iUnit,LCD
c
      write(iUnit,100) nSym
      do i=1,nSym
         call Write_Line_Symmetry(iUnit,Sym(1,1,i),LCD,LOGF)
      end do
c
 100  format(1x,i10,' # of symmetry elements in group')
c
      return
      end
c
c
      subroutine Read_Symmetry(iUnit,Gen)
      implicit none
c
      integer i,iUnit,j,Gen(3,5)
c
      read(iUnit,*) ((Gen(i,j),j=1,3),i=1,3) ! Rotation Matrix
      read(iUnit,*) ((Gen(i,j),j=4,5),i=1,3) ! Translation Vector
c
      return
      end
c
c
      subroutine Write_Line_Symmetry(iUnit,Gen,LCD,LOGF)
      implicit none
c
      integer i,iUnit,j,Gen(3,4),LCD,list(2,3),n,d,LOGF
c
      do i=1,3
         n=Gen(i,4)
         d=LCD
         call Fraction_Reduce(n,d)
         list(1,i)=n
         list(2,i)=d
      end do
c
      if(iUnit.eq.0) then
         write(LOGF,12) ((Gen(i,j),j=1,3),i=1,3),
     $        ((list(i,j),i=1,2),j=1,3)
      else
         write(iUnit,12) ((Gen(i,j),j=1,3),i=1,3),((list(i,j),i=1,2),
     $        j=1,3)
      end if
c
 12   format(3(3(1x,i2),3x),2x,3(2(1x,i3),2x))
c
      return
      end
c
c
      subroutine Read_Group(iUnit,nGen,Gen,nGenMax,nSymUnit,cellType)
      implicit none
c
      integer i,iUnit,nGen,nGenMax,nSymUnit
      integer Gen(3,5,nGenMax)
      character*3 cellType
c
      read(iUnit,*) nGen,nSymUnit
      read(iUnit,*) cellType
      do i=1,nGen
         call Read_Symmetry(iUnit,Gen(1,1,i))
      end do
c
      return
      end
c
c
      subroutine Get_SpaceGroup_FileName(iUnit,SpaceGroup,
     $     SpaceGroupFile,l,DEBUG,LOGF)
      implicit none
c
      integer i,iUnit,l,l1,SpaceGroup(2),LOGF
      character*3 SpaceGroupIndx,SpaceGroupConv
      character*50 SpaceGroupPath,SpaceGroupFile
      logical DEBUG
c
      if(DEBUG) call DEBUG_OC(0,'Get_SpaceGroup_FileName',23)
c
c -- Path to space group database from working directory
      read(iUnit,'(a)') SpaceGroupPath 
c
c -- Determine length of string containing space group path
      l=0
      do i=1,50
         if(SpaceGroupPath(i:i).eq.' ') then 
            l=i-1
            goto 90
         end if
      end do
 90   continue
c
      if(DEBUG) then 
         write(LOGF,*)
         write(LOGF,*) 'Path to SpaceGroup database: ',
     $     SpaceGroupPath(1:l)
      end if
c
c -- Convert space group name from integer to string
      call INTEGERtoSTRING(SpaceGroup(1),SpaceGroupIndx)
      call INTEGERtoSTRING(SpaceGroup(2),SpaceGroupConv)
c
c -- Concatenate path and filename
      SpaceGroupFile=SpaceGroupPath(1:l)
      if(SpaceGroupFile(l:l).ne.'/') then 
         SpaceGroupFile=SpaceGroupFile(1:l) // '/'
         l=l+1
      end if
c
c -- Length of File Name Suffix
      if(SpaceGroupConv(1:2).eq.'00') then
         l1=3
      elseif(SpaceGroupConv(1:1).eq.'0') then
         l1=2
      else
         l1=1
      end if
c
c -- Complete space group file name
      SpaceGroupFile=SpaceGroupFile(1:l) // SpaceGroupIndx //
     $     '.' // SpaceGroupConv(l1:3)
      l=l+3+2+(3-l1)
c
      if(l.gt.50) then
         write(LOGF,*) 'Error: Get_SpaceGroup_FileName '
         write(LOGF,*) 'Path to space group file has too many',
     $        ' characters'
         write(LOGF,*) 'Expecting length maximum of 50'
         stop
      end if
c
      if(DEBUG) then 
         write(LOGF,*) 'Space Group read from file:  ',
     $        SpaceGroupIndx//'.'//SpaceGroupConv(l1:3)
         write(LOGF,*) 'Full path to file:           ',
     $        SpaceGroupFile(1:l)
         call DEBUG_OC(1,'Get_SpaceGroup_FileName',23)
      end if
c     
      return
      end
c
c
      subroutine Read_Line_Bond(iUnit,bond)
      implicit none
c
      integer i,iUnit,j,k,bond(3,2,2)
c
      read(iUnit,*) (((bond(k,j,i),j=1,2),k=1,3),i=1,2)
c
      return
      end
c
c
      subroutine Read_Bonds(InputFile,fl,LCD_GEN)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer fl,i,i1,iUnit,j,k,l,denom(nBondsUnitMax*6),LCD_GEN,
     $     Bonds_Temp(3,2,2,nBondsUnitMax)
      character*50 InputFile
c
      if(DEBUG) call DEBUG_OC(0,'Read_Bonds',10)
c
      call Check_File(InputFile(1:fl),fl,LOGF)
c
      iUnit=1
      open(unit=iUnit,file=InputFile(1:fl))
      read(iUnit,*) nBondsUnit
c
c -- Enough memory allocated?
      if(nBondsUnit.gt.nBondsSimMax) then
         write(LOGF,*)
         write(LOGF,*) 'Error: Read_Bonds'
         write(LOGF,*) 'Not enough memory allocated'
         write(LOGF,*) 'Requested(nBondsUnit): ',nBondsUnit
         write(LOGF,*) 'Allowed(nBondsSimMax): ',nBondsSimMax
         stop
      end if
c
      do i=1,nBondsUnit
         call Read_Line_Bond(iUnit,Bonds_Temp(1,1,1,i))
      end do
      close(1)
c
c -- Determine LCD for vertices and bonds.
      l=0
      do i=1,nBondsUnit
         do j=1,2
            do k=1,3
               l=l+1
               denom(l)=Bonds_Temp(k,2,j,i)
            end do
         end do
      end do
      l=l+1
      denom(l)=LCD_GEN          ! Include LCD from generators
      call Lowest_Common_Multiple(denom,l,LCD_VS)
c
      if(DEBUG) then 
         write(LOGF,*)
         write(LOGF,101) LCD_VS
      end if
c
c -- Scale bonds by LCD_VS
      do i=1,nBondsUnit
         do j=1,2
            do k=1,3
               i1=LCD_VS/Bonds_Temp(k,2,j,i)
               HBonds(k,j,i)=Bonds_Temp(k,1,j,i)*i1
            end do
         end do
      end do
c
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,102) nBondsUnit
         call DEBUG_OC(1,'Read_Bonds',10)
      end if
c
 101  format(1x,'LCD for vertices, bonds, symmetry operations: ',i10)
 102  format(1x,'# of H-bonds defining unit cell: ',13x,i10)
c
      return
      end
c
c
      subroutine Read_Asym(InputFile,fl,LCD_GEN)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer fl,i,i1,iUnit,j,k,l,denom(nVertUnitMax*3),LCD_GEN,
     $     Vert_Temp(3,2,nVertUnitMax)
      character*50 InputFile
c
      if(DEBUG) call DEBUG_OC(0,'Read_Asym',9) 
c
      call Check_File(InputFile(1:fl),fl,LOGF)
c
      iUnit=1
      open(unit=iUnit,file=InputFile(1:fl))
      read(iUnit,*) nVertAsym
c
c -- Enough memory allocated?
      if(nVertAsym.gt.nVertUnitMax) then
         write(LOGF,*)
         write(LOGF,*) 'Error: Read_Bonds'
         write(LOGF,*) 'Not enough memory allocated'
         write(LOGF,*) 'Requested(nVertAsym): ',nVertAsym
         write(LOGF,*) 'Allowed(nVertUnitMax): ',nVertUnitMax
         stop
      end if
c
      do i=1,nVertAsym
         read(iUnit,*) ((Vert_Temp(k,j,i),j=1,2),k=1,3)
      end do
      close(1)
c
c -- Determine LCD for vertices
      l=0
      do i=1,nVertAsym
         do k=1,3
            l=l+1
            denom(l)=Vert_Temp(k,2,i)
         end do
      end do
      l=l+1
      denom(l)=LCD_GEN          ! Include LCD from generators
      call Lowest_Common_Multiple(denom,l,LCD_VS)
c
      if(DEBUG) then 
         write(LOGF,*)
         write(LOGF,*) 'LCD for vertices, bonds, symmetry operations: ',
     $        LCD_VS
      end if
c
c -- Scale vertices by LCD_VS
      do i=1,nVertAsym
         do k=1,3
            i1=LCD_VS/Vert_Temp(k,2,i)
            VertAsym(k,i)=Vert_Temp(k,1,i)*i1
         end do
      end do
c
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) '# of vertices in asymmetric unit: ',nVertAsym
         write(LOGF,*) 'Vertices(real*8)'
         do i=1,nVertAsym
            write(LOGF,10) (dfloat(VertAsym(j,i))/dfloat(LCD_VS),j=1,3)
         end do
      end if
c
      if(DEBUG) call DEBUG_OC(1,'Read_Asym',9) 
c
 10   format(2(5x,3(f8.5,1x)))
c
      return
      end
c
c
      subroutine Write_XYZ(TYPE,OutputFile,fl)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer i,j,k,iUnit,fl,l,nAtoms,nOXY,nHYD,tL
      real*8 r(3),Cell_Bound(3,8),OXY(3,nVertSimMax),
     $     HYD(3,nBondsSimMax),LV(3,3),LVI(3,3)
c$$$      real*8 r1(3)
      character*5 TYPE
      character*50 OutputFile,title
c
c -- The string TYPE is a five letter code describing which geometry to 
c -- write to file.
c -- 
c -- 1: U(Unit Cell) // S(Simulation) // C(Covering) // N(Non-Primitive)
c -- 2: O(Oxygen coordinates)
c -- 3: H(Hydrogen coordinates)
c -- 4: M(Hyd at midpoint of Bond) // B(Use bond variables)
c -- 5: P(Wrap Hyd within cell boundaries)
      if(DEBUG) write(LOGF,*) 'Writing ',OutputFile(1:fl)
c
c -- Initialize
      nAtoms=0
c
c -- Draw cell boundaries of unit cell?
      if(CELLBOUNDARY) then
         nAtoms=nAtoms+8
         l=0
         do i=0,1
            r(1)=dfloat(i)
            do j=0,1
               r(2)=dfloat(j)
               do k=0,1
                  r(3)=dfloat(k)
                  l=l+1
                  if((.not.PRIMITIVE).or.(TYPE(1:1).eq.'N')) then
                     call Matrix_Multiply_vM(r,LV_Unit_NonPrim,
     $                    Cell_Bound(1,l))
                  else
                     call Matrix_Multiply_vM(r,LV_Unit,Cell_Bound(1,l))
                  end if
               end do
            end do
         end do
      end if
c
c -- Which cell?
      if(TYPE(1:1).eq.'U') then
         nOXY=nVertUnit
         nHYD=nBondsUnit
         do i=1,3
            do j=1,3
               if(PRIMITIVE) then
                  LV(i,j)=LV_Unit(i,j)
                  LVI(i,j)=LVInv_Unit(i,j)
               else
                  LV(i,j)=LV_Unit_NonPrim(i,j)
                  LVI(i,j)=LVInv_Unit_NonPrim(i,j)
               end if
            end do
         end do
      elseif(TYPE(1:1).eq.'S') then
         nOXY=nVertSim
         nHYD=nBondsSim
         do i=1,3
            do j=1,3
               if(PRIMITIVE) then
                  LV(i,j)=LV_Sim(i,j)
                  LVI(i,j)=LVInv_Sim(i,j)
               else
                  LV(i,j)=LV_Sim_NonPrim(i,j)
                  LVI(i,j)=LVInv_Sim_NonPrim(i,j)
               end if
            end do
         end do
      elseif(TYPE(1:1).eq.'C') then
         nOXY=nVertCov
         nHYD=nBondsCov
         do i=1,3
            do j=1,3
               if(PRIMITIVE) then
                  LV(i,j)=LV_Cov(i,j)
                  LVI(i,j)=LVInv_Cov(i,j)
               else
                  LV(i,j)=LV_Cov_NonPrim(i,j)
                  LVI(i,j)=LVInv_Cov_NonPrim(i,j)
               end if
            end do
         end do
      elseif(TYPE(1:1).eq.'N') then
         nOXY=nVertC
         nHYD=nBondsC
         do i=1,3
            do j=1,3
               LV(i,j)=LV_Unit_NonPrim(i,j)
               LVI(i,j)=LVInv_Unit_NonPrim(i,j)
            end do
         end do
      else
         write(LOGF,*) 'Error: Write_XYZ'
         write(LOGF,*) 'Unknown Option(TYPE): ',TYPE
         stop
      end if
c
c -- Generate oxygen coordinates
      if(TYPE(2:2).eq.'O') then
         nAtoms=nAtoms+nOXY
         do i=1,nOXY
            if(TYPE(1:1).eq.'N') then
               r(1)=dfloat(VerticesC(1,i))/dfloat(LCD_VS)
               r(2)=dfloat(VerticesC(2,i))/dfloat(LCD_VS)
               r(3)=dfloat(VerticesC(3,i))/dfloat(LCD_VS)
            else
               r(1)=dfloat(Vertices(1,i))/dfloat(LCD_VS)
               r(2)=dfloat(Vertices(2,i))/dfloat(LCD_VS)
               r(3)=dfloat(Vertices(3,i))/dfloat(LCD_VS)
            end if
            call Matrix_Multiply_vM(r,LV_Unit,OXY(1,i))
         end do
      end if
c
c -- Generate hydrogen coordinates
      if(TYPE(3:3).eq.'H') then
         nAtoms=nAtoms+nHYD
         if(TYPE(4:4).eq.'M') then
            do i=1,nHYD
               if(TYPE(1:1).eq.'N') then
                  r(1)=dfloat(HBondsC(1,1,i)+HBondsC(1,2,i))/
     $                 dfloat(2*LCD_VS)
                  r(2)=dfloat(HBondsC(2,1,i)+HBondsC(2,2,i))/
     $                 dfloat(2*LCD_VS)
                  r(3)=dfloat(HBondsC(3,1,i)+HBondsC(3,2,i))/
     $                 dfloat(2*LCD_VS)
               else
                  r(1)=dfloat(HBonds(1,1,i)+HBonds(1,2,i))/
     $                 dfloat(2*LCD_VS)
                  r(2)=dfloat(HBonds(2,1,i)+HBonds(2,2,i))/
     $                 dfloat(2*LCD_VS)
                  r(3)=dfloat(HBonds(3,1,i)+HBonds(3,2,i))/
     $                 dfloat(2*LCD_VS)
               end if
               call Matrix_Multiply_vM(r,LV_Unit,HYD(1,i))
            end do
c
         elseif(TYPE(4:4).eq.'B') then
            write(LOGF,*) 'Warning: Write_XYZ'
            write(LOGF,*) 'TYPE: ',TYPE
            write(LOGF,*) 'Not yet coded'
            stop
c
         else
            write(LOGF,*) 'Error: Write_XYZ'
            write(LOGF,*) 'Unknown Option(TYPE): ',TYPE
            stop
         end if
c     
         if(TYPE(5:5).eq.'P') then
            do i=1,nHYD
               r(1)=HYD(1,i)
               r(2)=HYD(2,i)
               r(3)=HYD(3,i)
               call rToPrim(r,LV,LVI,HYD(1,i))
            end do
         end if
      end if
c
c -- Generate title for xyz file
      if(TYPE.eq.'UO   ') then
         title='Unit Cell: Oxy'
         tL=14
      elseif(TYPE.eq.'UOHM ') then
         title='Unit Cell: Oxy + Hyd at midpoints'
         tL=33
      elseif(TYPE.eq.'UOHMP') then
         title='Unit Cell: Oxy + Hyd(wrap) at midpoints'
         tL=39
      elseif(TYPE.eq.'SO   ') then
         title='Simulation Cell: Oxy'
         tL=20
      elseif(TYPE.eq.'SOHM ') then
         title='Simulation Cell: Oxy + Hyd at midpoints'
         tL=39
      elseif(TYPE.eq.'SOHMP') then
         title='Simulation Cell: Oxy + Hyd(wrap) at midpoints'
         tL=45
      elseif(TYPE.eq.'CO   ') then
         title='Covering Cell: Oxy'
         tL=18
      elseif(TYPE.eq.'COHM ') then
         title='Covering Cell: Oxy + Hyd at midpoints'
         tL=37
      elseif(TYPE.eq.'COHMP') then
         title='Covering Cell: Oxy + Hyd(wrap) at midpoints'
         tL=43
      elseif(TYPE.eq.'NO   ') then
         title='Cover_NP: Oxy'
         tL=13
      elseif(TYPE.eq.'NOHM') then
         title='Cover_NP: Oxy + Hyd at midpoints'
         tL=32
      elseif(TYPE.eq.'NOHMP') then
         title='Cover_NP: Oxy + Hyd(wrap) at midpoints'
         tL=39
      else
         write(LOGF,*) 'Warning: Write_XYZ'
         write(LOGF,*) 'Unknown Option(TYPE): ',TYPE
         title=' '
         tL=1
      end if
c
c -- Write coordinates to file
      iUnit=1
      open(unit=iUnit,file=OutputFile(1:fl))
      write(iUnit,*) nAtoms
      write(iUnit,*) title(1:tL)
c
c -- Write oxygen coordinates to file
      if(TYPE(2:2).eq.'O') then
         do i=1,nOXY
            write(iUnit,10) ' O',(OXY(j,i),j=1,3)
         end do
      end if
c
c -- Write hydrogen coordinates to file
      if(TYPE(3:3).eq.'H') then
         do i=1,nHYD            ! Oxygen Positions
            write(iUnit,10) ' H',(HYD(j,i),j=1,3)
         end do
      end if
c
c -- Write Cell Boundary to file?
      if(CELLBOUNDARY) then
         do i=1,8
            write(iUnit,10) ' X',(Cell_Bound(j,i),j=1,3)
         end do
      end if
c
      close(iUnit)
c
 10   format(2x,a2,3(3x,f11.6))
c
      return
      end
c
c
      subroutine Write_Structure(TYPE,OutputFile,fL)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer d,i,iUnit,j,j1,j2,k,n,fL,list(12),nOXY,nHYD
      character*1 TYPE
      character*50 OutputFile
c
      if(DEBUG) then 
         call DEBUG_OC(0,'Write_Structure',15)
         write(LOGF,*)
         write(LOGF,*) 'Writing structure to file: ',OutputFile(1:fL)
      end if
c
c -- The string TYPE determines for which cell structure information
c -- is written to file.
c
c -- Open output file
      iUnit=1
      open(unit=iUnit,file=OutputFile(1:fL))
c
c -- Write lattice vectors of primitive unit cell to file.  These are
c -- the vectors used to generate coordinates.
      write(iUnit,9) (LV_Unit(1,j),j=1,3),
     $           'Primitive Unit Cell Lattice Vectors'
      write(iUnit,10) (LV_Unit(2,j),j=1,3)
      write(iUnit,10) (LV_Unit(3,j),j=1,3)
c
c -- Write lattice vectors of simulation cell to file.  These are 
c -- the vectors used for periodic boundary conditions.
c
      if(TYPE.eq.'U') then      ! Unit Cell
         nOXY=nVertUnit
         nHYD=nBondsUnit
         if(PRIMITIVE) then
            write(iUnit,9) (LV_Unit(1,j),j=1,3),'Lattice Vectors'
            write(iUnit,10) (LV_Unit(2,j),j=1,3)
            write(iUnit,10) (LV_Unit(3,j),j=1,3)
         else
            write(iUnit,9) (LV_Unit_NonPrim(1,j),j=1,3),
     $           'Lattice Vectors'
            write(iUnit,10) (LV_Unit_NonPrim(2,j),j=1,3)
            write(iUnit,10) (LV_Unit_NonPrim(3,j),j=1,3)
         end if
      elseif(TYPE.eq.'S') then  ! Simulation Cell
         nOXY=nVertSim
         nHYD=nBondsSim
         if(PRIMITIVE) then
            write(iUnit,9) (LV_Sim(1,j),j=1,3),'Lattice Vectors'
            write(iUnit,10) (LV_Sim(2,j),j=1,3)
            write(iUnit,10) (LV_Sim(3,j),j=1,3)
         else
            write(iUnit,9) (LV_Sim_NonPrim(1,j),j=1,3),
     $           'Lattice Vectors'
            write(iUnit,10) (LV_Sim_NonPrim(2,j),j=1,3)
            write(iUnit,10) (LV_Sim_NonPrim(3,j),j=1,3)
         end if
      elseif(TYPE.eq.'C') then  ! Covering Cell
         nOXY=nVertCov
         nHYD=nBondsCov
         if(PRIMITIVE) then
            write(iUnit,9) (LV_Cov(1,j),j=1,3),'Lattice Vectors'
            write(iUnit,10) (LV_Cov(2,j),j=1,3)
            write(iUnit,10) (LV_Cov(3,j),j=1,3)
         else
            write(iUnit,9) (LV_Cov_NonPrim(1,j),j=1,3),
     $           'Lattice Vectors'
            write(iUnit,10) (LV_Cov_NonPrim(2,j),j=1,3)
            write(iUnit,10) (LV_Cov_NonPrim(3,j),j=1,3)
         end if
      elseif(TYPE.eq.'N') then  ! Non-primitive
         nOxy=nVertC
         nHYD=nBondsC
         write(iUnit,9) (LV_Unit_NonPrim(1,j),j=1,3),
     $        'Lattice Vectors'
         write(iUnit,10) (LV_Unit_NonPrim(2,j),j=1,3)
         write(iUnit,10) (LV_Unit_NonPrim(3,j),j=1,3)
      else
         write(LOGF,*) 'Error: Write_Structure'
         write(LOGF,*) 'Unknown Option(TYPE): ',TYPE
         write(LOGF,*) 'Expecting: U, S, or C'
         stop
      end if
c
c -- Write vertices to file
      write(iUnit,11) nOXY
      do i=1,nOXY
         k=0
         do j1=1,3
            if(TYPE.eq.'N') then
               n=VerticesC(j1,i)
            else
               n=Vertices(j1,i)
            end if
            d=LCD_VS
            call Fraction_Reduce(n,d)
            list(k+1)=n
            list(k+2)=d
            k=k+2
         end do
         write(iUnit,12) (list(j),j=1,6)
      end do
c
c -- Write HBonds to file
      write(iUnit,13) nHYD
      do i=1,nHYD
         k=0
         do j1=1,2
            do j2=1,3
               if(TYPE.eq.'N') then
                  n=HBondsC(j2,j1,i)
               else
                  n=HBonds(j2,j1,i)
               end if
               d=LCD_VS
               call Fraction_Reduce(n,d)
               list(k+1)=n
               list(k+2)=d
               k=k+2
            end do
         end do
c
         if(TYPE.eq.'U') then
            write(iUnit,14) (list(j),j=1,12),(BondListUnit(j,i),j=1,2)
         elseif(TYPE.eq.'S') then
            write(iUnit,14) (list(j),j=1,12),(BondListSim(j,i),j=1,2)
         elseif(TYPE.eq.'C') then
            write(iUnit,14) (list(j),j=1,12),(BondListCov(j,i),j=1,2)
         elseif(TYPE.eq.'N') then
            write(iUnit,14) (list(j),j=1,12),(BondListC(j,i),j=1,2)
         end if
      end do
c
      close(iUnit)
c
      if(DEBUG) call DEBUG_OC(1,'Write_Structure',15)
c
  9   format(1x,3(2x,f10.6),3x,35a)
 10   format(1x,3(2x,f10.6))
 11   format(1x,i5,'   # of vertices')
 12   format(1x,3(2(i5,1x),3x))
 13   format(1x,i5,'   # of HBonds')
 14   format(1x,3(2(i5,1x),3x),3x,3(2(i5,1x),3x),2(1x,i6))
c
      return
      end
c
c
      subroutine Write_GraphInvar(InputFile,fL)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer i,iUnit,j,k,fL
      character*50 InputFile
c
c -- Open file
      iUnit=1
      open(unit=iUnit,file=InputFile(1:fL))
c
      write(iUnit,10) nInv
      do i=1,nInv
         write(iUnit,11) GraphInvar_Type(i)
         write(iUnit,12) GraphInvar_Pair(1,i),GraphInvar_Pair(2,i),
     $        GraphInvar_nTerms(i),GraphInvar_Norm(i)
         do j=1,GraphInvar_nTerms(i)
            write(iUnit,13) (GraphInvar(k,j,i),k=1,3)
         end do
      end do
c
      close(iUnit)
c
 10   format(1x,i10,27x,'# of Invariants')
 11   format(1x,i1,36x,'0: b[j],   1:b[j]b[k],   2:b[j]c[k]')
 12   format(4(1x,i6),10x,'Bond Pair, # of Terms & Normalization')
 13   format(1x,i3,2(1x,i6))
c
      return
      end
c
c
      subroutine Write_GenBondPair(InputFile1,fL1,InputFile2,fL2)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer i,iUnit,iUnit2,j,k,l,nI,fL1,fL2,BOND(2,3,2),n,d,i1,i2
      real*8 dist,list(nSym)
      character*50 InputFile1,InputFile2
c
c -- How many first order invariants?
      nI=0
      do i=1,nInv
         if(GraphInvar_Type(i).ne.0) goto 100
         nI=nI+1
      end do
 100  continue
c
c -- Open file for graph invariants
      iUnit=1
      open(unit=iUnit,file=InputFile1(1:fL1))
c
c -- Open file for bond pairs to be used as input a separate calculation
      iUnit2=2
      open(unit=iUnit2,file=InputFile2(1:fL2))
c
      write(iUnit2,*) nInv,'     # of invariants'
c
      write(iUnit,*) 'Total # of invariants generated:     ',nInv
      write(iUnit,*)
      write(iUnit,*) '# of 1st order invariants:     ',nI
      write(iUnit,*) '========================================='
      write(iUnit,*)
      if(nI.ne.0) then
         do i=1,nI            ! Loop over b[j] invariants
            write(iUnit2,*) GraphInvar_Type(i),' : Type'
c
            write(iUnit,*)
            write(iUnit,*) 'Invariant: ',i,',  Type: ',
     $           GraphInvar_Type(i)
            write(iUnit,*) 'Generating Bonds: '
c
            do j=1,nBondPairs(i)
c
c -- Write bonds as rational numbers
               do i1=1,2
                  do i2=1,3
                     n=HBonds(i2,i1,BondPair(1,j,i))
                     d=LCD_VS
                     call Fraction_Reduce(n,d)
                     BOND(1,i2,i1)=n
                     BOND(2,i2,i1)=d
                  end do
               end do
c
               write(iUnit,10) j,BondPair(1,j,i),
     $              (((BOND(k,i2,i1),k=1,2),i2=1,3),i1=1,2)
               if(j.eq.1) then 
                  write(iUnit2,9) (((BOND(k,i2,i1),k=1,2),i2=1,3),
     $                 i1=1,2)
               end if
            end do
         end do
      end if
c
      write(iUnit,*)
      write(iUnit,*) '# of 2nd order invariants: ',nInv-nI
      write(iUnit,*) '========================================='
      write(iUnit,*)
      do i=nI+1,nInv          ! Loop over 2nd order invariants
         write(iUnit2,*) GraphInvar_Type(i),' : Type'
c
         write(iUnit,*)
         write(iUnit,*) 'Invariant: ',i,',  Type: ',GraphInvar_Type(i)
c
         do j=1,nBondPairs(i)
            call BondPair_MinDist(BondPair(1,j,i),BondPair(2,j,i),
     $           list(j))
         end do
         call List_Min_Real(list,nBondPairs(i),dist)
         write(iUnit,*) 'Minimum Distance Generating Bond Pairs: '
c
         l=0
         do j=1,nBondPairs(i)
            if(abs(list(j)-dist).le.TOL_ZERO) then 
               l=l+1
c
c -- Write 1st bond as rational number
               do i1=1,2
                  do i2=1,3
                     n=HBonds(i2,i1,BondPair(1,j,i))
                     d=LCD_VS
                     call Fraction_Reduce(n,d)
                     BOND(1,i2,i1)=n
                     BOND(2,i2,i1)=d
                  end do
               end do
c
c -- Write the first nBondPairsMax pairs to file
               write(iUnit,11) l,BondPair(1,j,i),list(j),
     $              (((BOND(k,i2,i1),k=1,2),i2=1,3),i1=1,2)
c
c -- Write the first pair to file ready for a second calculation.
               if(l.eq.1) then 
                  write(iUnit2,9) (((BOND(k,i2,i1),k=1,2),i2=1,3),
     $                 i1=1,2)
               end if
c
c -- Write 1st bond as rational number
               do i1=1,2
                  do i2=1,3
                     n=HBonds(i2,i1,BondPair(2,j,i))
                     d=LCD_VS
                     call Fraction_Reduce(n,d)
                     BOND(1,i2,i1)=n
                     BOND(2,i2,i1)=d
                  end do
               end do
c
c -- Write the first nBondPairsMax pairs to file
               write(iUnit,12) BondPair(2,j,i),
     $              (((BOND(k,i2,i1),k=1,2),i2=1,3),i1=1,2)
c
c -- Write the first pair to file ready for a second calculation.
               if(l.eq.1) then 
                  write(iUnit2,9) (((BOND(k,i2,i1),k=1,2),i2=1,3),
     $                 i1=1,2)
               end if
            end if
         end do
      end do
c
      close(iUnit)
      close(iUnit2)
c
 9    format(1x,                    3(3x,2(i5,1x)),5x,3(3x,2(i5,1x)))
 10   format(1x,i5,' : ',i5,11x,    3(3x,2(i5,1x)),5x,3(3x,2(i5,1x)))
 11   format(1x,i5,' : ',i5,4x,f7.3,3(3x,2(i5,1x)),5x,3(3x,2(i5,1x)))
 12   format(1x,8x,i5,11x,          3(3x,2(i5,1x)),5x,3(3x,2(i5,1x)))
c
      return
      end
c
c
      subroutine Read_BondPair(InputFile,fL)
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer fL,i,i1,iV,j,j1,j2,k,iUnit,list(12),n,d
      character*50 InputFile
c
      if(DEBUG) call DEBUG_OC(0,'Read_BondPairs',14)
c
      call Check_File(InputFile(1:fL),fL,LOGF)
c
      iUnit=1
      open(unit=iUnit,file=InputFile(1:fL))
c
      read(iUnit,*) nInv
      do i=1,nInv
         iV=0
         read(iUnit,*) k
         GraphInvar_Type(i)=k
c
 100     continue
         iV=iV+1
         read(iUnit,*) (list(j),j=1,12) ! Read bonds as rational numbers
c
c -- Scale bonds by LCD_VS
         do j1=1,2
            do j2=1,3
               n=list((j1-1)*6+(j2-1)*2+1)
               d=list((j1-1)*6+(j2-1)*2+2)
               i1=LCD_VS/d
               if(i1*d.ne.LCD_VS) then
                  write(LOGF,*) 'Error: Read_BondPair'
                  write(LOGF,*) 'The denominator of the bond',
     $                 ' coordinates'
                  write(LOGF,*) 'does not divide into LCD_VS an integer'
                  write(LOGF,*) 'number of times.'
                  write(LOGF,*) 'Bond Pair: ',i
                  write(LOGF,*) 'LCD_VS: ',LCD_VS
                  write(LOGF,*) 'Denominator: ',d
                  stop
               end if
               GenBond(j2,j1,iV,i)=n*i1
            end do
         end do
         if((GraphInvar_Type(i).gt.0).and.(iV.lt.2)) goto 100 ! Read 2nd bond
      end do
c
      close(iUnit)
c
      if(DEBUG) call DEBUG_OC(1,'Read_BondPairs',14)
c
      return
      end
c
c
      subroutine Write_Memory(nI,nR,name,nL,LOGF)
      implicit none
c
      integer i,nI,nR,nL,LOGF
      real*8 tI(4),tR(4)
      character*50 name
c
      tI(1)=dfloat(nI*4)
      tR(1)=dfloat(nR*8)
      do i=2,4
         tI(i)=tI(i-1)/dfloat(1024)
         tR(i)=tR(i-1)/dfloat(1024)
      end do
      write(LOGF,11) name(1:nL),(nI*4)+(nR*8),(tI(i)+tR(i),i=2,4)
c
 11   format(a28,1x,i13,' Bytes,',1x,f11.2,' KB,',1x,f8.2,' MB,',1x,
     $     f5.2,' GB')
c
      return
      end
c
c
      subroutine Write_Enumeration_Input()
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
      include 'graphinvar.inc'
c
      integer i,j,r(3),n(2),list(nBondsSimMax),FindVertex,
     $     Bond(3,2),IB(2),nV,nS,nL
      integer nB
c
      if(DEBUG) then 
         call DEBUG_OC(0,'Write_Enumeration_Input',23)
         write(LOGF,*)
         write(LOGF,*) 'Writing file Bonds...'
      end if
c
c -- Write vertex indices for bonds
      open(unit=1,file='Bonds') 
      if(COVERCELL) then
         nV=nVertCov
         nB=nBondsCov
         nS=nSym
         do i=1,nB
            write(1,*) (BondListCov(j,i),j=1,2)
         end do
      elseif(SIMCELL) then
         nV=nVertSim
         nB=nBondsSim
         nS=nSym
         do i=1,nB
            write(1,*) (BondListSim(j,i),j=1,2)
         end do
      else
         nV=nVertUnit
         nB=nBondsUnit
         nS=nSymUnit
         do i=1,nB
            write(1,*) (BondListUnit(j,i),j=1,2)
         end do
      end if
      close(1)
c
c -- Write permutation group of vertices
      if(DEBUG) write(LOGF,*) 'Writing file Gv...'
      open(unit=1,file='Gv')
      do j=1,nS
         do i=1,nV
            if(COVERCELL) then
               call Sym_OPP(Sym(1,1,j),Vertices(1,i),r,Cell_Cov,
     $              CellInv_Cov)
            elseif(SIMCELL) then
               call Sym_OPP(Sym(1,1,j),Vertices(1,i),r,Cell_Sim,
     $              CellInv_Sim)
            else
               call Sym_OPP(Sym(1,1,j),Vertices(1,i),r,Cell,CellInv)
            end if
            list(i)=FindVertex(r,nV,Vertices)
         end do
         write(1,9) (list(i),i=1,nV)
      end do
      close(1)
c
c -- Write permutation group of bonds
      if(DEBUG) write(LOGF,*) 'Writing file Gbonds...'
      open(unit=1,file='Gbonds')
      do j=1,nS
         do i=1,nB
            call Sym_OPBOND(Sym(1,1,j),HBonds(1,1,i),BOND)
            if(COVERCELL) then
               call Sym_IBOND(BOND,Cell_Cov,CellInv_Cov,nB,IB)
            elseif(SIMCELL) then
               call Sym_IBOND(BOND,Cell_Sim,CellInv_Sim,nB,IB)
            else
               call Sym_IBOND(BOND,Cell,CellInv,nB,IB)
            end if
            list(i)=IB(1)*IB(2)
         end do
         write(1,9) (list(i),i=1,nB)
      end do
      close(1)
c
c -- # of b[j]b[k] invariants and total # of terms
      n(1)=0
      n(2)=0
      do i=1,nInv
         if(GraphInvar_Type(i).eq.1) then
            n(1)=n(1)+1
            n(2)=n(2)+GraphInvar_nTerms(i)
         end if
      end do
c
c -- Write invariants to file(only b[j]b[k] type)
      if(DEBUG) write(LOGF,*) 'Writing file Invar...'
      open(unit=1,file='Invar')
      write(1,*) n(1),n(2)
      do i=1,nInv
         if(GraphInvar_Type(i).eq.1) then
            write(1,*) GraphInvar_nTerms(i)
            do j=1,GraphInvar_nTerms(i)
               write(1,*) GraphInvar(2,j,i),GraphInvar(3,j,i),
     $              GraphInvar(1,j,i)
            end do
         end if
      end do
      close(1)
c
c -- Write bond constraints if covering cell was used
c -- Loop over covering cell bonds only.  Wrap each bond back into the 
c -- simulation cell and identify index of bond.
      if(COVERCELL) then
         if(DEBUG) write(LOGF,*) 'Writing file BondsC...'
         open(unit=1,file='BondsC')
         write(1,*) nBondsSim,'   # of Groups'
         do i=1,nBondsSim       ! Loop over bonds in simulation cell
            nL=0
            do j=1,nBondsCov
               if(BondMap(j).eq.i) then
                  nL=nL+1
                  list(nL)=j
               end if
            end do
            write(1,*) nL,'   # of bonds in Group ',i
            do j=1,nL
               write(1,*) list(j)
            end do
         end do
         close(1)
      elseif(COVERNP) then
         if(DEBUG) write(LOGF,*) 'Writing file BondsC...'
         open(unit=1,file='BondsC')
         write(1,*) nBondsC,'   # of Groups'
         do i=1,nBondsC         ! Loop over bonds in simulation cell
            nL=0
            do j=1,nBondsSim
               if(BondMap(j).eq.i) then
                  nL=nL+1
                  list(nL)=j
               end if
            end do
            write(1,*) nL,'   # of bonds in Group ',i
            do j=1,nL
               write(1,*) list(j)
            end do
         end do
         close(1)
      end if
c
      if(DEBUG) call DEBUG_OC(1,'Write_Enumeration_Input',23)
c
 9    format(1000(i4,1x))
c
      return
      end
