!#########################################################################
!  Structure Generation Routines for MkInvar: 
!     GenStructure                 -- Generate structure for all cells.
!     Generate_Vertices_HBonds     -- Use H-bond definitions to generate list 
!                                      of vertices.
!     Generate_Vertices_Asym       -- Use vertices of asymmetric unit to 
!                                      generate all vertices possible in the 
!                                      unit cell.
!     Generate_HBonds_Asym         -- Identify all H-Bonds possible in the unit
!                                      cell.
!     Vert_Compare                 -- Compare vertice with list of vertices
!                                      test if unique.
!     Generate_LatticeVector       -- Use six lattice parameters to generate
!                                      lattice vectors for unit cell.
!     Check_LatticeParam           -- Check that six lattice parameters and 
!                                      cell type for space group are consistent.
!     Orbit_Vertices               -- Identify orbits for vertices
!     Orbit_HBonds                 -- Identify orbits for bonds
!     Generate_Vertices_HBonds_Sim -- Generate xyz coordinates for simulation 
!                                      cell.
!     Generate_Vertices_HBonds_Cov -- "      "     " for covering cell.
!     CoverNP_Vertices_HBonds      -- Identify vertices and bonds for non-
!                                      primitive cell from those in covering
!                                      cell.
!     Generate_BondMap             -- Generate bond map for covering cell
!                                      calculations.
!     
!
!#########################################################################   
      subroutine GenStructure()
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
c
c -- Generate structure for unit cell
      if(jobHBonds.eq.0) then   ! Generate H-Bonds from assymetric unit
         call Generate_Vertices_Asym()
         call Generate_HBonds_Asym()
      elseif(jobHBonds.eq.1) then ! Read H-Bonds from file
         call Generate_Vertices_HBonds()
      else                      ! Invalid Option
         write(LOGF,*) 
         write(LOGF,*) 'Error: MkInvar'
         write(LOGF,*) 'Invalid option for determing H-bonds',
     $        '(jobHBonds): ',jobHBonds
         stop
      end if
c
      if(DEBUG) then 
         write(LOGF,*)
         write(LOGF,*) 'TERMINATE: ',TERMINATE
      end if
c
c -- Write unit cell geometries to file
      call Write_XYZ('UO   ','Vert_Unit.xyz',13) ! Oxy
      call Write_XYZ('UOHM ','Bond_Unit.xyz',13) ! Oxy + Hyd at midpoint
      call Write_XYZ('UOHMP','Bond_Unit_Wrap.xyz',18) ! Oxy + Hyd(wrap) at midpoint
c
c -- By this point, the vertices and bonds that define the unit cell
c -- are known.
      call Write_Structure('U','Structure_Unit.txt',18)
c
c -- We will terminate the program if the oxygen atoms are not all four 
c -- coordinated.  Information about orbits is helpful when deciding
c -- which bonds to remove from list.
      if(TERMINATE) call EXIT_TERMINATE(1,LOGF)
c
c -- Test if vertices and bonds form sets of orbits
      call Orbit_Vertices(nVertUnit,nSymUnit)
      call Orbit_HBonds(nBondsUnit,nSymUnit)
c
c -- We terminate the program at this point if the vertices and bonds
c -- do not form complete orbits.  Any invariant generated would just 
c -- be garbage.
      if(TERMINATE) then 
         call EXIT_TERMINATE(2,LOGF)
         stop
      end if
c
c -- Write Simulation cell geometries to file
      if(SIMCELL) then
         call Generate_Vertices_HBonds_Sim()
         call Write_XYZ('SO   ','Vert_Sim.xyz',12) ! Oxy
         call Write_XYZ('SOHM ','Bond_Sim.xyz',12) ! Oxy + Hyd at midpoint
         call Write_XYZ('SOHMP','Bond_Sim_Wrap.xyz',17)      ! Oxy + Hyd(wrap) at midpoint
         call Write_Structure('S','Structure_Sim.txt',17)
      end if
c
c -- Write Covering cell geometries to file
      if(COVERCELL) then 
         call Generate_Vertices_HBonds_Cov()
         call Write_XYZ('CO   ','Vert_Cov.xyz',12) ! Oxy
         call Write_XYZ('COHM ','Bond_Cov.xyz',12) ! Oxy + Hyd at midpoint
         call Write_XYZ('COHMP','Bond_Cov_Wrap.xyz',17)      ! Oxy + Hyd(wrap) at midpoint
         call Write_Structure('C','Structure_Cov.txt',17)
      end if
c
c -- If covering cell is used for calculations on a non-primitive cell, then 
c -- find those bonds and vertices contained in the non-primitive cell.
      if(COVERNP) then
         call CoverNP_Vertices_HBonds()
         call Write_XYZ('NO   ','VertC.xyz',9)
         call Write_XYZ('NOHM ','BondC.xyz',9)
         call Write_XYZ('NOHMP','BondC_Wrap.xyz',14)
         call Write_Structure('N','StructureC.txt',14)
      end if
c
c -- Generate bond map for covering cell calculations
      if(COVERCELL.or.COVERNP) call Generate_BondMap()
c
      return
      end
c
c
      subroutine Generate_Vertices_HBonds()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
c
      integer i,j,iT,i1,i2,i3,j1,k,cOXY(nVertUnitMax),v(3),IB(2),
     $     FindVertex,T(3),Bond(3,2),HBonds_Temp(3,2,nBondsUnitMax)
      real*8 r(3),r1(3),OXY(3,nVertUnitMax),dist,FUNC_NORM
      logical test
c
c -- The convention for the definition of H-bonds is that the first 
c -- vertex in each pair lie within the unit cell.
      if(DEBUG) then
         call DEBUG_OC(0,'Generate_Vertices_HBonds',24)
         write(LOGF,*) 'Using H-Bonds to identify unique vertices'
      end if
c
c -- If non-primitive cell, apply translation group to unit cell
c -- bonds to generate all bonds in cell.
      if(.not.PRIMITIVE) then
         if(DEBUG) then 
            write(LOGF,*)
            write(LOGF,*) 'Generating Bonds for non-primitive cell'
            write(LOGF,*) '  Translating unit cell bonds'
         end if
c
c -- Return unit cell bonds to non-primitive cell
         do i=1,nBondsUnit
            do j=1,2
               do k=1,3
                  HBonds_Temp(k,j,i)=HBonds(k,j,i)
               end do
            end do
         end do
c
         iT=1
         call Sym_PRIM(HBonds_Temp(1,1,1),Cell,CellInv,HBonds(1,1,iT))
         do i=2,nBondsUnit
            call Sym_PRIM(HBonds_Temp(1,1,i),Cell,CellInv,
     $           HBonds(1,1,iT+1)) ! Return to unit cell
            call FindBond(HBonds(1,1,iT+1),iT,IB,HBonds)
            if(IB(2).eq.0) iT=iT+1 ! Found new bond?  
         end do
         nBondsUnit=iT
c
         if(DEBUG) then
            write(LOGF,*) '# of bonds in non-primitive cell: ',
     $           nBondsUnit
            write(LOGF,*) 
            write(LOGF,*) '  Applying translation group to bonds.'
         end if
c
c -- Loop over all translations contained in covering cell
         iT=nBondsUnit
         do i1=NonPrimRange(1,1),NonPrimRange(2,1) ! Translations along lattice vector #1
            T(1)=i1*LCD_VS
            do i2=NonPrimRange(1,2),NonPrimRange(2,2) ! Translations along #2
               T(2)=i2*LCD_VS
               do i3=NonPrimRange(1,3),NonPrimRange(2,3) ! Translations along #3
                  if((i1.eq.i2).and.(i2.eq.i3).and.(i3.eq.0)) goto 100 ! skip over unit cell at origin
                  T(3)=i3*LCD_VS
c     
c -- Loop over symmetry operations of unit cell
                  do i=1,nBondsUnit ! Loop over unit cell bonds
                     Bond(1,1)=HBonds(1,1,i)+T(1) ! Add translation
                     Bond(2,1)=HBonds(2,1,i)+T(2)
                     Bond(3,1)=HBonds(3,1,i)+T(3)
c
                     Bond(1,2)=HBonds(1,2,i)+T(1)
                     Bond(2,2)=HBonds(2,2,i)+T(2)
                     Bond(3,2)=HBonds(3,2,i)+T(3)
c
                     call Sym_PRIM(Bond,Cell,CellInv,HBonds(1,1,iT+1)) ! Return to unit cell
                     call FindBond(HBonds(1,1,iT+1),iT,IB,HBonds)
                     if(IB(2).eq.0) iT=iT+1 ! Found new bond?
                  end do
c     
c -- We skip over unit cell at origin
 100              continue
               end do
            end do
         end do
c
         nBondsUnit=iT
         if(DEBUG) write(LOGF,*) '# of Bonds in non-primitive cell: ',
     $        nBondsUnit
      end if
c
c -- Initialize list of vertices
      nVertUnit=1
      Vertices(1,1)=HBonds(1,1,1) ! Denominators equal one
      Vertices(2,1)=HBonds(2,1,1)
      Vertices(3,1)=HBonds(3,1,1)
c
c -- Loop over all H-bonds and keep unique vertices
      do i=1,nBondsUnit
         do j=1,2
            call rToPrim_Integer(HBonds(1,j,i),Cell,CellInv,
     $           Vertices(1,nVertUnit+1))
            call Vert_Compare(nVertUnit+1,test)
            if(test) nVertUnit=nVertUnit+1
c
c -- Enough memory allocated?
            if(nVertUnit.gt.nVertUnitMax) then
               write(LOGF,*)
               write(LOGF,*) 'Error: Generate_Vertices_HBonds'
               write(LOGF,*) 'Not enough memory allocated'
               write(LOGF,*) 'Requested(nVertUnit):   ',nVertUnit
               write(LOGF,*) 'Allowed(nVertUnitMax):  ',nVertUnitMax
               write(LOGF,*) '# of Bonds(nBondsUnit): ',nBondsUnit
               stop
            end if
         end do
      end do
c
c -- Initialize coordination numbers
      do i=1,nVertUnit
         cOXY(i)=0              ! Coordination # for oxygen
      end do
c     
c -- Generate physical coordinates of oxygen atoms
      if(DEBUG) then 
         write(LOGF,*)
         write(LOGF,*) 'Physical Oxygen Coordinates'
      end if
c
      do i=1,nVertUnit
         r(1)=dfloat(Vertices(1,i))/dfloat(LCD_VS)
         r(2)=dfloat(Vertices(2,i))/dfloat(LCD_VS)
         r(3)=dfloat(Vertices(3,i))/dfloat(LCD_VS)
         call Matrix_Multiply_vM(r,LV_Unit,OXY(1,i))
c$$$         r(1)=dfloat(Vertices(1,i))
c$$$         r(2)=dfloat(Vertices(2,i))
c$$$         r(3)=dfloat(Vertices(3,i))
c$$$         call Matrix_Multiply_vM(r,CellInv,r1)
c$$$         call Matrix_Multiply_vM(r1,LV_Unit,OXY(1,i))
         if(DEBUG) write(LOGF,10) i,(OXY(j,i),j=1,3)
      end do
c
c -- Loop over pairs of oxygens and test if below allowed minimum
c -- distance
      do i=1,nVertUnit-1
         do j=i+1,nVertUnit
            if(PRIMITIVE) then 
               call MinImage(OXY(1,i),OXY(1,j),LV_Unit,r1,ORTHOG)
            else
               call MinImage(OXY(1,i),OXY(1,j),LV_Unit_NonPrim,r1,
     $              ORTHOG)
            end if
            dist=FUNC_NORM(r1)
            if(dist.lt.TOL_VV) then
               TERMINATE=.true.
               write(LOGF,*)
               write(LOGF,*) 'Warning: Generate_Vertices_HBonds'
               write(LOGF,*) 'Oxygens too close(dist):      ',dist
               write(LOGF,*) 'Minimum allowed dist(TOL_VV): ',TOL_VV
               write(LOGF,*) 'r1: ',(r1(k),k=1,3)
               write(LOGF,*) 'i: ',i
               write(LOGF,*) (OXY(k,i),k=1,3)
               write(LOGF,*) 'j: ',j
               write(LOGF,*) (OXY(k,j),k=1,3)
            end if
         end do
      end do
c
c -- Loop over bonds and identify index of vertices
      do i=1,nBondsUnit
         do j=1,2
            call rToPrim_Integer(HBonds(1,j,i),Cell,CellInv,v)
            j1=FindVertex(v,nVertUnit,Vertices)
c
            if(j1.eq.0) then
               write(LOGF,*) 'Error: Generate_Vertices_HBonds'
               write(LOGF,*) 'Vertex not Found: ',(v(k),k=1,3)
               stop
            end if
            BondListUnit(j,i)=j1
            cOXY(j1)=cOXY(j1)+1
         end do
      end do
c
c -- Check oxygen coordination
      do i=1,nVertUnit
         if(cOXY(i).ne.4) TERMINATE=.true.
      end do
c
      if(DEBUG.or.TERMINATE) then 
         write(LOGF,*)
         write(LOGF,101) TOL_HBOND(1)
         write(LOGF,102) TOL_HBOND(2)
         write(LOGF,*)
         write(LOGF,*) '# of oxygens: ',nVertUnit
         write(LOGF,11) 'Index','x','y','z','Coord.'
         write(LOGF,'(1x,40(1h-))')
         do i=1,nVertUnit
            write(LOGF,10) i,
     $           (dfloat(Vertices(j,i))/dfloat(LCD_VS),j=1,3),cOXY(i)
         end do
         write(LOGF,*)
         write(LOGF,*) '# of H-Bonds Found: ',nBondsUnit
         write(LOGF,12) 'Vertex 1','Vertex 2','Oxygen'
         write(LOGF,13) 'Index','x','y','z','x','y','z','Indices'
         write(LOGF,'(1x,66(1h-))')
         do i=1,nBondsUnit
            write(LOGF,14) i,
     $           (dfloat(HBonds(j,1,i))/dfloat(LCD_VS),j=1,3),
     $           (dfloat(HBonds(j,2,i))/dfloat(LCD_VS),j=1,3),
     $           (BondListUnit(j,i),j=1,2)
         end do            
      end if
c
      if(DEBUG) call DEBUG_OC(1,'Generate_Vertices_HBonds',24)
c
 10   format(1x,i3,3(3x,f6.2),5x,i2)
 11   format(1x,a5,5x,a1,8x,a1,8x,a1,5x,a6)
 12   format(14x,a8,19x,a8,11x,a6)
 13   format(1x,a5,4x,a1,7x,a1,7x,a1,10x,a1,7x,a1,7x,a1,6x,a7)
 14   format(1x,i3,3(2x,f6.2),3x,3(2x,f6.2),4x,i2,2x,i2)
c
 101  format(1x,'Min H-Bond Distance Tolerance: ',f12.6)
 102  format(1x,'Max H-Bond Distance Tolerance: ',f12.6)
c
      return
      end
c
c
      subroutine Generate_Vertices_Asym()
      implicit none
c
      include 'param.inc'
      include 'symmetry.inc'
      include 'structure.inc'
c
      integer i,j,k
      real*8 r(3),r1(3),OXY(3,nVertUnitMax),dist,FUNC_NORM
      logical test
c
      if(DEBUG) call DEBUG_OC(0,'Generate_Vertices_Asym',22)
c
c -- Initialize list of vertices
c -- Loop over vertices of assymetric unit & symmetery elements in group
      nVertUnit=0
      do i=1,nVertAsym
         do j=1,nSymUnit
            call Sym_OPP(Sym(1,1,j),VertAsym(1,i),
     $           Vertices(1,nVertUnit+1),Cell,CellInv)
            call Vert_Compare(nVertUnit+1,test)
            if(test) nVertUnit=nVertUnit+1
         end do
         if(DEBUG) write(LOGF,*) 'Asym Vert: ',i,' # of vertices: ',
     $        nVertUnit
      end do
c
c -- Enough memory?
      if(nVertUnit.gt.nVertUnitMax) then
            write(LOGF,*)
            write(LOGF,*) 'Error: Generate_Vertices_Asym'
            write(LOGF,*) 'Not enough memory allocated'
            write(LOGF,*) 'Requested(nVertUnit):   ',nVertUnit
            write(LOGF,*) 'Allowed(nVertUnitMax):  ',nVertUnitMax
            stop
      end if
c
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) '# of unique vertices in unit cell: ',nVertUnit
      end if
c     
c -- Generate physical coordinates of oxygen atoms
      do i=1,nVertUnit
         r(1)=dfloat(Vertices(1,i))/dfloat(LCD_VS)
         r(2)=dfloat(Vertices(2,i))/dfloat(LCD_VS)
         r(3)=dfloat(Vertices(3,i))/dfloat(LCD_VS)
         call Matrix_Multiply_vM(r,LV_Unit,OXY(1,i))
c$$$         r(1)=dfloat(Vertices(1,i))
c$$$         r(2)=dfloat(Vertices(2,i))
c$$$         r(3)=dfloat(Vertices(3,i))
c$$$         call Matrix_Multiply_vM(r,CellInv,r1)
c$$$         call Matrix_Multiply_vM(r1,LV_Unit,OXY(1,i))
      end do
c
c -- Loop over pairs of oxygens and test if below allowed minimum
c -- distance
      do i=1,nVertUnit-1
         do j=i+1,nVertUnit
            if(PRIMITIVE) then
               call MinImage(OXY(1,i),OXY(1,j),LV_Unit,r1,ORTHOG)
            else
               call MinImage(OXY(1,i),OXY(1,j),LV_Unit_NonPrim,r1,
     $              ORTHOG)
            end if
            dist=FUNC_NORM(r1)
            if(dist.lt.TOL_VV) then
               TERMINATE=.true.
               write(LOGF,*) 'Warning: Generate_Vertices_HBonds'
               write(LOGF,*) 'Oxygens too close(dist): ',dist
               write(LOGF,*) 'r1: ',(r1(k),k=1,3)
               write(LOGF,*) 'i: ',i
               write(LOGF,*) (OXY(k,i),k=1,3)
               write(LOGF,*) 'j: ',j
               write(LOGF,*) (OXY(k,j),k=1,3)
            end if
         end do
      end do
c
      if(DEBUG) call DEBUG_OC(1,'Generate_Vertices_Asym',22)
c
      return
      end
c
c
      subroutine Generate_HBonds_Asym()
      implicit none
c
      include 'param.inc'
      include 'structure.inc' 
c
      integer i,i1,i2,j,j1,j2,j3,iT,cOXY(nVertUnitMax),TRANS_INT(3,27)
      real*8 r(3),r1(3),OXY(3,nVertUnitMax),TRANS(3,27),dist,FUNC_NORM
c
      if(DEBUG) call DEBUG_OC(0,'Generate_HBonds_Asym',20)
c
c -- Initialization
      nBondsUnit=0              ! # of HBonds found
      do i=1,nVertUnit
         cOXY(i)=0              ! Coordination # for oxygen
      end do
c
      iT=0
      do j1=-1,1                ! Loop over lattice translations
         r(1)=dfloat(j1)
         do j2=-1,1 
            r(2)=dfloat(j2)
            do j3=-1,1
               r(3)=dfloat(j3)
               iT=iT+1
               if(PRIMITIVE) then
                  call Matrix_Multiply_vM(r,LV_Unit,TRANS(1,iT)) ! real*8 translation
               else
                  call Matrix_Multiply_vM(r,LV_Unit_NonPrim,TRANS(1,iT)) ! real*8 translation
               end if
               call Matrix_Multiply_vM(r,Cell,r1) ! integer translation
               TRANS_INT(1,iT)=int(r1(1)*TOL_INTEGER)
               TRANS_INT(2,iT)=int(r1(2)*TOL_INTEGER)
               TRANS_INT(3,iT)=int(r1(3)*TOL_INTEGER)
c$$$               TRANS_INT(1,iT)=j1*LCD_VS ! integer translation
c$$$               TRANS_INT(2,iT)=j2*LCD_VS
c$$$               TRANS_INT(3,iT)=j3*LCD_VS
            end do
         end do
      end do      
c     
c -- Generate physical coordinates of oxygen atoms
      do i=1,nVertUnit
         r(1)=dfloat(Vertices(1,i))/dfloat(LCD_VS)
         r(2)=dfloat(Vertices(2,i))/dfloat(LCD_VS)
         r(3)=dfloat(Vertices(3,i))/dfloat(LCD_VS)
         call Matrix_Multiply_vM(r,LV_Unit,OXY(1,i))
c$$$         r(1)=dfloat(Vertices(1,i))
c$$$         r(2)=dfloat(Vertices(2,i))
c$$$         r(3)=dfloat(Vertices(3,i))
c$$$         call Matrix_Multiply_vM(r,CellInv,r1)
c$$$         call Matrix_Multiply_vM(r1,LV_Unit,OXY(1,i))
      end do
c
c -- For every unique pair of oxygens, loop over all periodic images.  
c -- If distance between two oxygen atoms is within tolerance, then 
c -- note them as possibly forming a hydrogen bond.
      do i1=1,nVertUnit-1
         do i2=i1+1,nVertUnit
c
            do iT=1,27
               r(1)=OXY(1,i1)-(OXY(1,i2)+TRANS(1,iT))
               r(2)=OXY(2,i1)-(OXY(2,i2)+TRANS(2,iT))
               r(3)=OXY(3,i1)-(OXY(3,i2)+TRANS(3,iT))
               dist=FUNC_NORM(r)
c
c -- Possible H-Bond found.  Keep track of oxygen coordination number.
c -- H-Bonds are stored in their integer form instead of converting 
c -- real*8 to integer later on.
               if((dist.gt.TOL_HBOND(1))
     $              .and.(dist.le.TOL_HBOND(2))) then
                  cOXY(i1)=cOXY(i1)+1
                  cOXY(i2)=cOXY(i2)+1
                  nBondsUnit=nBondsUnit+1
                  BondListUnit(1,nBondsUnit)=i1
                  BondListUnit(2,nBondsUnit)=i2
c
c -- Enough memory allocated
                  if(nBondsUnit.gt.nBondsSimMax) then
                     write(LOGF,*) 'Error: Generate_HBonds_Asym'
                     write(LOGF,*) 'Not enough memory allocated'
                     write(LOGF,*) '# of HBonds found(nBondsUnit): ',
     $                    nBondsUnit
                     write(LOGF,*) 'Maximum allowed(nBondsSimMax): ',
     $                    nBondsSimMax
                     write(LOGF,*) 'Reasonable guess(8*nVertUnit): ',
     $                    8*nVertUnit
                     stop
                  end if
c
c -- Numerator for Vertex 1 of H-Bond
                  HBonds(1,1,nBondsUnit)=Vertices(1,i1)
                  HBonds(2,1,nBondsUnit)=Vertices(2,i1)
                  HBonds(3,1,nBondsUnit)=Vertices(3,i1)
c
c -- Numerator for Vertex 2 of H-Bond
                  HBonds(1,2,nBondsUnit)=Vertices(1,i2)+TRANS_INT(1,iT)
                  HBonds(2,2,nBondsUnit)=Vertices(2,i2)+TRANS_INT(2,iT)
                  HBonds(3,2,nBondsUnit)=Vertices(3,i2)+TRANS_INT(3,iT)
               end if
            end do
c
        end do
      end do
c
c -- Check oxygen coordination
      do i=1,nVertUnit
         if(cOXY(i).ne.4) TERMINATE=.true.
      end do
c
      if(DEBUG.or.TERMINATE) then 
         write(LOGF,*)
         write(LOGF,*) 'Min H-Bond Distance Tolerance: ',TOL_HBOND(1)
         write(LOGF,*) 'Max H-Bond Distance Tolerance: ',TOL_HBOND(2)
         write(LOGF,*)
         write(LOGF,*) '# of oxygens: ',nVertUnit
         write(LOGF,11) 'Index','x','y','z','Coord.'
         write(LOGF,'(1x,47(1h-))')
         do i=1,nVertUnit
            write(LOGF,10) i,
     $           (dfloat(Vertices(j,i))/dfloat(LCD_VS),j=1,3),cOXY(i)
         end do
         write(LOGF,*)
         write(LOGF,*) '# of H-Bonds: ',nBondsUnit
         write(LOGF,12) 'Vertex 1','Vertex 2','Oxygen'
         write(LOGF,13) 'Index','x','y','z','x','y','z','Indices'
         write(LOGF,'(1x,80(1h-))')
         do i=1,nBondsUnit
            write(LOGF,14) i,
     $           (dfloat(HBonds(j,1,i))/dfloat(LCD_VS),j=1,3),
     $           (dfloat(HBonds(j,2,i))/dfloat(LCD_VS),j=1,3),
     $           (BondListUnit(j,i),j=1,2)
         end do         
      end if
c
      if(DEBUG) call DEBUG_OC(1,'Generate_HBonds_Asym',20)
c
 10   format(1x,i3,3(3x,f8.5),5x,i4)
 11   format(1x,a5,5x,a1,10x,a1,10x,a1,8x,a6)
 12   format(17x,a8,25x,a8,16x,a6)
 13   format(1x,a5,4x,a1,9x,a1,9x,a1,12x,a1,9x,a1,9x,a1,10x,a7)
 14   format(1x,i3,3(2x,f8.5),3x,3(2x,f8.5),4x,i4,2x,i4)
c
      return
      end
c
c
      subroutine Vert_Compare(n,test)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer i,j1,k,n
      logical test
c
c -- Is the last element already present in the list?
      do i=1,n-1
         do k=1,3
            j1=abs(Vertices(k,i)-Vertices(k,n))
            if(j1.gt.0) goto 91
         end do
         goto 90                ! Vertices are identical
c     
 91      continue               ! Vertices are different
      end do
c     
      test=.true.               ! The vertex is unique to this list
      return
c
 90   continue
      test=.false.              ! The vertex is already present in this list
c
      return
      end
c
c
      subroutine Generate_LatticeVector()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      real*8 d,e,f,alpha,beta,gamma
c
      if(DEBUG) call DEBUG_OC(0,'Generate_LatticeVector',22)
c
      d=PI/180.d0               ! Convert decimal to radians
      alpha=LV_Param(4)*d
      beta= LV_Param(5)*d
      gamma=LV_Param(6)*d
c
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) 'Angles(Radians): '
         write(LOGF,101) alpha
         write(LOGF,102) beta
         write(LOGF,103) gamma
         write(LOGF,*)
         write(LOGF,104) dcos(alpha)
         write(LOGF,*)
         write(LOGF,105) dsin(beta)
         write(LOGF,106) dcos(beta)
         write(LOGF,*)
         write(LOGF,107) dsin(gamma)
         write(LOGF,108) dcos(gamma)
         write(LOGF,109) dtan(gamma)
      end if
c
      e=dcos(alpha)/dsin(gamma)-dcos(beta)/dtan(gamma)
      f=dsqrt(dsin(beta)*dsin(beta)-e*e)
c
c -- Lattice vector one(LV1) is parallel to x-axis
      LV_Unit(1,1)=LV_Param(1)
      LV_Unit(1,2)=0.d0
      LV_Unit(1,3)=0.d0
c
c -- Lattice vector two(LV2) is contained in the xy-plane and 
c -- makes an angle \gamma with LV1
      LV_Unit(2,1)=LV_Param(2)*dcos(gamma)
      LV_Unit(2,2)=LV_Param(2)*dsin(gamma)
      LV_Unit(2,3)=0.d0
c
c -- LV3 may have nonzero xyz components and makes angles \alpha and
c -- \beta with LV2 and LV1 respectively.
      LV_Unit(3,1)=LV_Param(3)*dcos(beta)
      LV_Unit(3,2)=LV_Param(3)*e
      LV_Unit(3,3)=LV_Param(3)*f
c
      if(DEBUG) call DEBUG_OC(1,'Generate_LatticeVector',22)
c
 101  format(1x,'Alpha:',6x,f13.6)
 102  format(1x,'Beta:',7x,f13.6)
 103  format(1x,'Gamma:',6x,f13.6)
 104  format(1x,'Cos(alpha): ',f13.6)
 105  format(1x,'Sin(beta):  ',f13.6)
 106  format(1x,'Cos(beta):  ',f13.6)
 107  format(1x,'Sin(gamma): ',f13.6)
 108  format(1x,'Cos(gamma): ',f13.6)
 109  format(1x,'Tan(gamma): ',E13.6E2)
c
      return
      end
c
c
      subroutine Check_LatticeParam(CT)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      character*3 CT,TEST

      logical AB_SAME,AC_SAME,BC_SAME,ABC_SAME,A_90,B_90,G_90,G_60,
     $     G_120,AB_90,AG_90,ABG_90,ABG_SAME,ERROR
c
      if(DEBUG) call DEBUG_OC(0,'Check_LatticeParam',18)
c
      ERROR=.false.
      TEST='NOT'
      ORTHOG=.false.
c
c -- How are lengths and angles related?
      AB_SAME=.false.
      if(abs(LV_Param(1)-LV_Param(2)).le.TOL_ZERO) AB_SAME=.true.
c
      AC_SAME=.false.
      if(abs(LV_Param(1)-LV_Param(3)).le.TOL_ZERO) AC_SAME=.true.
c
      BC_SAME=.false.
      if(abs(LV_Param(2)-LV_Param(3)).le.TOL_ZERO) BC_SAME=.true.
c
      ABC_SAME=.false.
      if(AB_SAME.and.AC_SAME) ABC_SAME=.true.  
c
      A_90=.false.
      if(abs(LV_Param(4)-90.d0).le.TOL_ZERO) A_90=.true.
c
      B_90=.false.
      if(abs(LV_Param(5)-90.d0).le.TOL_ZERO) B_90=.true.
c
      G_90=.false.
      if(abs(LV_Param(6)-90.d0).le.TOL_ZERO) G_90=.true.
c
      G_60=.false.
      if(abs(LV_Param(6)-60.d0).le.TOL_ZERO) G_60=.true.
c
      G_120=.false.
      if(abs(LV_Param(6)-120.d0).le.TOL_ZERO) G_120=.true.
c
      AB_90=.false.
      if(A_90.and.B_90) AB_90=.true.
c
      AG_90=.false.
      if(A_90.and.G_90) AG_90=.true.
c
      ABG_90=.false.
      if(AB_90.and.G_90) ABG_90=.true.
c
      ABG_SAME=.false.
      if((abs(LV_Param(4)-LV_Param(5)).le.TOL_ZERO)
     $     .and.(abs(LV_Param(4)-LV_Param(6)).le.TOL_ZERO)) then 
         ABG_SAME=.true.
      end if
c
c -- Use lattice parameters to determine cell type.  Does this agree with 
c -- cell type specified in space group file?
      if(ABC_SAME) then         ! a == b == c
         if(ABG_SAME) then ! alpha == beta == gamma
            if(ABG_90) then ! alpha == beta == gamma == 90
               TEST='cub'
               ORTHOG=.true.
            else
               TEST='trg'
               ORTHOG=.true.
            end if
         end if
      elseif(AB_SAME.and.(.not.AC_SAME)) then ! a == b != c
         if(ABG_90) then ! alpha == beta == gamma == 90
            TEST='tet'
            ORTHOG=.true.
         elseif(AB_90.and.(G_60.or.G_120)) then ! alpha == beta == 90, gamma == 60
            TEST='hex'
         end if
      elseif((.not.AB_SAME).and.(.not.BC_SAME).and.(.not.AC_SAME)) then ! a != b != c
         if(ABG_90) then 
            TEST='ort'
            ORTHOG=.true.
         elseif(AG_90.and.(.not.B_90)) then
            TEST='mon'
         elseif((.not.A_90).and.(.not.B_90)
     $           .and.(.not.G_90)) then
            TEST='trc'
         end if            
      end if
c
c -- Error detected in the lattice parameters
c      ERROR=.true.
      if(CT.ne.TEST) ERROR=.true.
c
c -- Lattice parameters are consistent with cell type
      if(DEBUG.or.ERROR) then
        write(LOGF,*) 
        write(LOGF,101) 
        write(LOGF,102) LV_Param(1)
        write(LOGF,103) LV_Param(2)
        write(LOGF,104) LV_Param(3)
        write(LOGF,105) LV_Param(4)
        write(LOGF,106) LV_Param(5)
        write(LOGF,107) LV_Param(6)
        write(LOGF,*) 
        write(LOGF,108) AB_SAME
        write(LOGF,109) AC_SAME
        write(LOGF,110) BC_SAME
        write(LOGF,111) ABC_SAME
        write(LOGF,112) A_90
        write(LOGF,113) B_90
        write(LOGF,114) G_90
        write(LOGF,115) G_60
        write(LOGF,116) G_120
        write(LOGF,117) AB_90
        write(LOGF,118) AG_90
        write(LOGF,119) ABG_SAME
        write(LOGF,120) ABG_90
        write(LOGF,*)
        write(LOGF,121) TEST
        write(LOGF,122) ORTHOG
        if(.not.PRIMITIVE) then
           ORTHOG=.false.
           write(LOGF,*) 
           write(LOGF,*) 'Non-Primitive cell vectors are treated as',
     $          ' non-orthogonal.'
           write(LOGF,*) 'ORTHOG set to                 ',ORTHOG
        end if
      end if
c
      if(ERROR) then
         write(LOGF,*) 'Error: Check_LatticeParam'
         write(LOGF,*) 'Lattice parameters inconsistent with cell',
     $        ' type: ',CT
         stop
      end if
c
      if(DEBUG) call DEBUG_OC(1,'Check_LatticeParam',18)
c
 101  format(1x,'Lattice Constants')
 102  format(1x,'Length of vector 1(a):',14x,f12.6)
 103  format(1x,'Length of vector 2(b):',14x,f12.6)
 104  format(1x,'Length of vector 3(c):',14x,f12.6)
 105  format(1x,'Angle between vectors 2 & 3(Alpha): ',f12.6)
 106  format(1x,'Angle between vectors 1 & 3(Beta):  ',f12.6)
 107  format(1x,'Angle between vectors 1 & 2(Gamma): ',f12.6)
 108  format(1x,'a == b?',24x,l1)
 109  format(1x,'a == c?',24x,l1)
 110  format(1x,'b == c?',24x,l1)
 111  format(1x,'a == b == c?',19x,l1)
 112  format(1x,'Alpha == 90?',19x,l1)
 113  format(1x,'Beta  == 90?',19x,l1)
 114  format(1x,'Gamma == 90?',19x,l1)
 115  format(1x,'Gamma == 60?',19x,l1)
 116  format(1x,'Gamma == 120?',18x,l1)
 117  format(1x,'Alpha == Beta == 90?',11x,l1)
 118  format(1x,'Alpha == Gamma == 90?',10x,l1)
 119  format(1x,'Alpha == Beta == Gamma?',8x,l1)
 120  format(1x,'Alpha == Beta == Gamma == 90?',2x,l1)
 121  format(1x,'Cell Type Found:',15x,3a)
 122  format(1x,'ORTHOGONAL(ORTHOG)?',12x,l1)
c
      return
      end
c
c
      subroutine Orbit_Vertices(nOXY,nOP)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
c
      integer i,i1,j,k,r(3),FindVertex,list(nSym),nL,nOXY,nOP
c
      if(DEBUG) then 
         call DEBUG_OC(0,'Orbit_Vertices',14)
         write(LOGF,*)
         write(LOGF,*) 'Vertex:     Symmetry Related Vertices'
      end if
c
c -- Loop over all vertices in unit cell and identify indices of 
c -- symmetry related vertices.
      do i=1,nOXY
         nL=nOP
         do j=1,nOP
            call Sym_OPP(Sym(1,1,j),Vertices(1,i),r,Cell,CellInv)
            i1=FindVertex(r,nOXY,Vertices)
            if(i1.eq.0) TERMINATE=.true.
            list(j)=i1
         end do
         call List_Union_Integer(list,nL)
         if(DEBUG) write(LOGF,10) i,(list(k),k=1,nL)
      end do
c
      if(DEBUG) call DEBUG_OC(1,'Orbit_Vertices',14)
c
 10   format(1x,i6,':',100(1x,i3))
c
      return
      end
c
c
      subroutine Orbit_HBonds(nBonds,nOP)
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
      include 'symmetry.inc'
c
      integer i,j,k,BOND(3,2),BOND2(3,2),IB(2),JB(2),nL,list(nSym),
     $     nBonds,nOP,nNewBonds
c
      if(DEBUG) then 
         call DEBUG_OC(0,'Orbit_HBonds',12)
         write(LOGF,*)
         write(LOGF,*) 'Bond  :     Symmetry Related Bonds'
      end if
c
      nNewBonds=0
      do i=1,nBonds
         nL=nOP
         do j=1,nOP
            call Sym_OPBOND(Sym(1,1,j),HBonds(1,1,i),BOND)
            call Sym_IBOND(BOND,Cell,CellInv,nBonds,IB)
            if(IB(2).eq.0) TERMINATE=.true.
            list(j)=IB(2)
c
c -- Keep track of new H-bonds
            if(IB(2).eq.0) then
               call Sym_PRIM(BOND,Cell,CellInv,HBonds(1,1,nBonds+1))
               write(LOGF,*)
               write(LOGF,*) 'HBond: ',i,'   Sym: ',j
               write(LOGF,*) 'New H-bond Identified'
               write(LOGF,*) 'Vertex 1: ',
     $              (HBonds(k,1,nBonds+1)/dfloat(LCD_VS),k=1,3)
               write(LOGF,*) 'Vertex 2: ',
     $              (HBonds(k,2,nBonds+1)/dfloat(LCD_VS),k=1,3)
               call FindBond(HBonds(1,1,nBonds+1),nBonds+nNewBonds,JB,
     $              HBonds)
               if(JB(2).eq.0) nNewBonds=nNewBonds+1
            end if
         end do
c
c -- List of unique bond indices in orbit
         call List_Union_Integer(list,nL)
         if(DEBUG) write(LOGF,10) i,(list(k),k=1,nL)
      end do
c
c -- Were there any new H-bonds identified?
      if(DEBUG) then 
         write(LOGF,*) '# of New H-bonds: ',nNewBonds
         do i=1,nNewBonds
            write(LOGF,*) i,': '
         end do
      end if
c
      if(DEBUG) call DEBUG_OC(1,'Orbit_HBonds',12)
c
 10   format(1x,i6,':',100(1x,i3))
c
      return
      end
c
c
      subroutine Generate_Vertices_HBonds_Sim()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer i,i1,i2,i3,j,j1,k,iV,iB,TRANS(3),v(3),FindVertex
      real*8 r(3),r1(3)
c
      if(DEBUG) call DEBUG_OC(0,'Generate_Vertices_HBonds_Sim',28)
c
c -- Enough memory allocated?
      nVertSim=nVertUnit*SimCellDim(1)*SimCellDim(2)*SimCellDim(3)
      if(nVertSim.gt.nVertSimMax) then
         write(LOGF,*)'Error: Generate_Vertices_HBonds_Sim'
         write(LOGF,*) 'Not enough memory allocated'
         write(LOGF,*) 'Requested(nVertSim):    ',nVertSim
         write(LOGF,*) 'Allowed(nVertSimMax):   ',nVertSimMax
         stop
      end if
c
      nBondsSim=nBondsUnit*SimCellDim(1)*SimCellDim(2)*SimCellDim(3)
      if(nBondsSim.gt.nBondsSimMax) then
         write(LOGF,*)'Error: Generate_Vertices_HBonds_Sim'
         write(LOGF,*) 'Not enough memory allocated'
         write(LOGF,*) 'Requested(nBondsSim):    ',nBondsSim
         write(LOGF,*) 'Allowed(nBondsSimMax):   ',nBondsSimMax
         stop
      end if
c
c -- Initialization
      do i=1,3
         do j=1,3
            Cell_Sim(i,j)=Cell(i,j)*dfloat(SimCellDim(i))
            LV_Sim(i,j)=LV_Unit(i,j)*dfloat(SimCellDim(i))
            if(.not.PRIMITIVE) LV_Sim_NonPrim(i,j)=LV_Unit_NonPrim(i,j)*
     $           dfloat(SimCellDim(i))
         end do
      end do
      call Matrix_Inverse(Cell_Sim,CellInv_Sim,LOGF)
      call Matrix_Inverse(LV_Sim,LVInv_Sim,LOGF)
      if(.not.PRIMITIVE) call Matrix_Inverse(LV_Sim_NonPrim,
     $     LVInv_Sim_NonPrim,LOGF)
c
c -- Loop over all unit cells in simulation cell
      iV=nVertUnit
      iB=nBondsUnit
      do i1=0,SimCellDim(1)-1
         r(1)=dfloat(i1)
         do i2=0,SimCellDim(2)-1
            r(2)=dfloat(i2)
            do i3=0,SimCellDim(3)-1
               r(3)=dfloat(i3)
               if((i1.eq.i2).and.(i2.eq.i3).and.(i3.eq.0)) goto 100 ! skip over unit cell at origin
c
c -- Translation vector
               call Matrix_Multiply_vM(r,Cell,r1)
               TRANS(1)=int(r1(1)*TOL_INTEGER)
               TRANS(2)=int(r1(2)*TOL_INTEGER)
               TRANS(3)=int(r1(3)*TOL_INTEGER)
c
c -- Loop over vertices in unit cell
               do i=1,nVertUnit
                  iV=iV+1
                  Vertices(1,iV)=Vertices(1,i)+TRANS(1)
                  Vertices(2,iV)=Vertices(2,i)+TRANS(2)
                  Vertices(3,iV)=Vertices(3,i)+TRANS(3)
               end do
c
c -- Loop over bonds in unit cell
               do i=1,nBondsUnit
                  iB=iB+1
                  HBonds(1,1,iB)=HBonds(1,1,i)+TRANS(1)
                  HBonds(2,1,iB)=HBonds(2,1,i)+TRANS(2)
                  HBonds(3,1,iB)=HBonds(3,1,i)+TRANS(3)
c
                  HBonds(1,2,iB)=HBonds(1,2,i)+TRANS(1)
                  HBonds(2,2,iB)=HBonds(2,2,i)+TRANS(2)
                  HBonds(3,2,iB)=HBonds(3,2,i)+TRANS(3)
               end do
c              
 100           continue
            end do
         end do
      end do
c
      if(DEBUG) then 
         write(LOGF,*) '# of vertices in Simulation Cell: ',nVertSim
         write(LOGF,*) '# of bonds in Simulation Cell:    ',nBondsSim
      end if
      if(iV.ne.nVertSim) then
         write(LOGF,*) 'Error: Generate_Vertices_HBonds_Sim'
         write(LOGF,*) 'Incorrect # of vertices'
         write(LOGF,*) '# of Vertices Found(iV):          ',iV
         write(LOGF,*) '# of Vertices Expected(nVertSim): ',nVertSim
         stop
      end if
c
      if(iB.ne.nBondsSim) then
         write(LOGF,*) 'Error: Generate_Vertices_HBonds_Sim'
         write(LOGF,*) 'Incorrect # of bonds'
         write(LOGF,*) '# of Bonds Found(iB):           ',iV
         write(LOGF,*) '# of Bonds Expected(nBondsSim): ',nVertSim
         stop
      end if
c
c -- Loop over bonds and identify index of vertices
      if(DEBUG) write(LOGF,*) ' Constructing BondListSim...'
      do i=1,nBondsSim
         do j=1,2
            call rToPrim_Integer(HBonds(1,j,i),Cell_Sim,CellInv_Sim,v)
            j1=FindVertex(v,nVertSim,Vertices)
c
            if(j1.eq.0) then
               write(LOGF,*) 'Error: Generate_Vertices_HBonds_Sim'
               write(LOGF,*) 'Vertex not found'
               write(LOGF,*) 'Bond: ',i,'   Vertex: ',j
               write(LOGF,*) 'Vertex(Before): ',(HBonds(k,j,i),k=1,3)
               write(LOGF,*) 'Vertex not Found: ',(v(k),k=1,3)
               stop
            end if
            BondListSim(j,i)=j1
         end do
      end do      
c
      if(DEBUG) call DEBUG_OC(1,'Generate_Vertices_HBonds_Sim',28)
c
      return
      end
c
c     
      subroutine Generate_Vertices_HBonds_Cov()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer i,i1,i2,i3,j,j1,k,iV,iB,TRANS(3),v(3),FindVertex
      real*8 r(3),r1(3)
      logical test(3)
c
      if(DEBUG) call DEBUG_OC(0,'Generate_Vertices_HBonds_Cov',28)
c
c -- Enough memory allocated?
      nVertCov=nVertUnit*(CovCellDim**3)
      if(nVertCov.gt.nVertSimMax) then
         write(LOGF,*)'Error: Generate_Vertices_HBonds_Cov'
         write(LOGF,*) 'Not enough memory allocated'
         write(LOGF,*) 'Requested(nVertCov):    ',nVertCov
         write(LOGF,*) 'Allowed(nVertSimMax):   ',nVertSimMax
         stop
      end if
c
      nBondsCov=nBondsUnit*(CovCellDim**3)
      if(nBondsCov.gt.nBondsSimMax) then
         write(LOGF,*)'Error: Generate_Vertices_HBonds_Cov'
         write(LOGF,*) 'Not enough memory allocated'
         write(LOGF,*) 'Requested(nBondsCov):    ',nBondsCov
         write(LOGF,*) 'Allowed(nBondsSimMax):   ',nBondsSimMax
         stop
      end if
c
c -- Initialization
      do i=1,3
         do j=1,3
            Cell_Cov(i,j)=Cell(i,j)*dfloat(CovCellDim)
            LV_Cov(i,j)=LV_Unit(i,j)*dfloat(CovCellDim)
            if(.not.PRIMITIVE) LV_Cov_NonPrim(i,j)=LV_Unit_NonPrim(i,j)*
     $           dfloat(CovCellDim)
         end do
      end do
      call Matrix_Inverse(Cell_Cov,CellInv_Cov,LOGF)
      call Matrix_Inverse(LV_Cov,LVInv_Cov,LOGF)
      if(.not.PRIMITIVE) call Matrix_Inverse(LV_Cov_NonPrim,
     $     LVInv_Cov_NonPrim,LOGF)
c
c -- Loop over all unit cells in covering cell and only add vertices
c -- and bonds to the lists if the unit cell is outside of the 
c -- simulation cell.
      iV=nVertSim
      iB=nBondsSim
      do i1=0,CovCellDim-1
         test(1)=.false.
         if(i1.gt.(SimCellDim(1)-1)) test(1)=.true.
         r(1)=dfloat(i1)
c
         do i2=0,CovCellDim-1
            test(2)=.false.
            if(i2.gt.(SimCellDim(2)-1)) test(2)=.true.
c
            r(2)=dfloat(i2)
            do i3=0,CovCellDim-1
               test(3)=.false.
               if(i3.gt.(SimCellDim(3)-1)) test(3)=.true.
c     
               if((i1.eq.i2).and.(i2.eq.i3).and.(i3.eq.0)) goto 100 ! skip over unit cell at origin
               r(3)=dfloat(i3)
c
c -- Translation vector
               if(test(1).or.test(2).or.test(3)) then
                  call Matrix_Multiply_vM(r,Cell,r1)
                  TRANS(1)=int(r1(1)*TOL_INTEGER)
                  TRANS(2)=int(r1(2)*TOL_INTEGER)
                  TRANS(3)=int(r1(3)*TOL_INTEGER)
c
c -- Loop over vertices in unit cell
                  do i=1,nVertUnit
                     iV=iV+1
                     Vertices(1,iV)=Vertices(1,i)+TRANS(1)
                     Vertices(2,iV)=Vertices(2,i)+TRANS(2)
                     Vertices(3,iV)=Vertices(3,i)+TRANS(3)
                  end do
c     
c -- Loop over bonds in unit cell
                  do i=1,nBondsUnit
                     iB=iB+1
                     HBonds(1,1,iB)=HBonds(1,1,i)+TRANS(1)
                     HBonds(2,1,iB)=HBonds(2,1,i)+TRANS(2)
                     HBonds(3,1,iB)=HBonds(3,1,i)+TRANS(3)
c     
                     HBonds(1,2,iB)=HBonds(1,2,i)+TRANS(1)
                     HBonds(2,2,iB)=HBonds(2,2,i)+TRANS(2)
                     HBonds(3,2,iB)=HBonds(3,2,i)+TRANS(3)
                  end do
               end if
c     
 100           continue
            end do
         end do
      end do
c
      if(DEBUG) then 
         write(LOGF,*) '# of vertices in Covering Cell: ',nVertCov
         write(LOGF,*) '# of bonds in Covering Cell:    ',nBondsCov
      end if
      if(iV.ne.nVertCov) then
         write(LOGF,*) 'Error: Generate_Vertices_HBonds_Cov'
         write(LOGF,*) 'Incorrect # of vertices'
         write(LOGF,*) '# of Vertices Found(iV):          ',iV
         write(LOGF,*) '# of Vertices Expected(nVertCov): ',nVertCov
         stop
      end if
c
      if(iB.ne.nBondsCov) then
         write(LOGF,*) 'Error: Generate_Vertices_HBonds_Cov'
         write(LOGF,*) 'Incorrect # of bonds'
         write(LOGF,*) '# of Bonds Found(iB):           ',iV
         write(LOGF,*) '# of Bonds Expected(nBondsCov): ',nVertCov
         stop
      end if
c
c -- Loop over bonds and identify index of vertices
      if(DEBUG) write(LOGF,*) ' Constructing BondListCov...'
      do i=1,nBondsCov
         do j=1,2
            call rToPrim_Integer(HBonds(1,j,i),Cell_Cov,CellInv_Cov,v)
            j1=FindVertex(v,nVertCov,Vertices)
c
            if(j1.eq.0) then
               write(LOGF,*) 'Error: Generate_Vertices_HBonds_Cov'
               write(LOGF,*) 'Vertex not found'
               write(LOGF,*) 'Bond: ',i,'   Vertex: ',j
               write(LOGF,*) 'Vertex(Before): ',(HBonds(k,j,i),k=1,3)
               write(LOGF,*) 'Vertex not Found: ',(v(k),k=1,3)
               stop
            end if
            BondListCov(j,i)=j1
         end do
      end do      
c
      if(DEBUG) call DEBUG_OC(1,'Generate_Vertices_HBonds_Cov',28)
c
      return
      end
c
c
      subroutine CoverNP_Vertices_HBonds()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer i,i1,i2,j,nB,nV,IB(2),HB(3,2),FindVertex
c
      if(DEBUG) call DEBUG_OC(0,'CoverNP_Vertices_HBonds',23)
c
c -- Loop over bonds in covering cell
      nB=0
      do i=1,nBondsSim
         do i2=1,2              ! Apply origin shift
            HB(1,i2)=HBonds(1,i2,i)
            HB(2,i2)=HBonds(2,i2,i)
            HB(3,i2)=HBonds(3,i2,i)
         end do
c
         call Sym_PRIM(HB,CellC,CellCInv,HBondsC(1,1,nB+1))
         if(i.eq.1) then
            nB=nB+1             ! First bond in list
         else
            call FindBond(HBondsC(1,1,nB+1),nB,IB,HBondsC)
            if(IB(2).eq.0) nB=nB+1 ! New bond
         end if
      end do
      nBondsC=nB
c
c -- Loop over bonds in non-primitive cell to find vertices
      nV=0
      do i=1,nBondsC
         do j=1,2
            call rToPrim_Integer(HBondsC(1,j,i),CellC,CellCInv,
     $           VerticesC(1,nV+1))
            if((i.eq.1).and.(j.eq.1)) then
               nV=nV+1
               BondListC(1,1)=1
            else
               i1=FindVertex(VerticesC(1,nV+1),nV,VerticesC)
               if(i1.eq.0) then 
                  nV=nV+1
                  BondListC(j,i)=nV
               else
                  BondListC(j,i)=i1
               end if               
            end if
         end do
      end do
      nVertC=nV
c
c -- List Vertices and Bonds
      if(DEBUG) then
         write(LOGF,*)
         write(LOGF,*) '# of Vertices in non-primitive cell: ',nV
         do i=1,nV
            write(LOGF,14) i,
     $           (dfloat(VerticesC(j,i))/dfloat(LCD_VS),j=1,3)
         end do
         write(LOGF,*)
         write(LOGF,*) '# of Bonds in non-primitive cell: ',nB
         do i=1,nB
            write(LOGF,14) i,
     $           (dfloat(HBondsC(j,1,i))/dfloat(LCD_VS),j=1,3),
     $           (dfloat(HBondsC(j,2,i))/dfloat(LCD_VS),j=1,3),
     $           (BondListC(j,i),j=1,2)
         end do
      end if
c
      if(DEBUG) call DEBUG_OC(1,'CoverNP_Vertices_HBonds',23)
c
 14   format(1x,i3,3(2x,f6.2),3x,3(2x,f6.2),4x,i2,2x,i2)
c
      return
      end
c
c
      subroutine Generate_BondMap()
      implicit none
c
      include 'param.inc'
      include 'structure.inc'
c
      integer i,IB(2),HB(3,2)
c
      if(DEBUG) call DEBUG_OC(0,'Generate_BondMap',16)
c
c -- Wrap each covering cell bond back into the simulation cell
c -- and identify index of bond.
      if(COVERCELL) then
         do i=0,nBondsSim       ! The case i=0 handles 1st order invariants
            BondMap(i)=i
         end do
         do i=nBondsSim+1,nBondsCov
            call Sym_IBOND(HBonds(1,1,i),Cell_Sim,CellInv_Sim,nBondsSim,
     $           IB)
            if((IB(1).eq.-1).or.(IB(2).eq.0)) then
               write(LOGF,*) 'Error: Generate_BondMap'
               write(LOGF,*) 'Problem with wrapping covering cell bond'
               write(LOGF,*) 'into simulation cell(COVERCELL).'
               write(LOGF,*) 'i: ',i
               write(LOGF,*) 'IB: ',IB(1),'  ',IB(2)
               stop
            end if
            BondMap(i)=IB(2)
         end do
c
c -- Wrap each covering cell bond back into the non-primitive cell
c -- and identify index of bond.
      elseif(COVERNP) then
         do i=1,nBondsSim
            call Sym_PRIM(HBonds(1,1,i),CellC,CellCInv,HB)
            call FindBond(HB,nBondsC,IB,HBondsC)
            if((IB(1).eq.-1).or.(IB(2).eq.0)) then
               write(LOGF,*) 'Error: Generate_BondMap'
               write(LOGF,*) 'Problem with wrapping covering cell bond'
               write(LOGF,*) 'into simulation cell(COVERNP).'
               write(LOGF,*) 'i: ',i
               write(LOGF,*) 'IB: ',IB(1),'  ',IB(2)
               stop
            end if
            BondMap(i)=IB(2)
         end do

      end if
c
      if(DEBUG) call DEBUG_OC(1,'Generate_BondMap',16)
c
      return
      end
